
; This attempts to remove the call and return protocols from PROC lambdas.
; If all of the calling points of a PROC lambda can be identified, REALLY-
; DETERMINE-PROTOCOL is called to do the actual work.

(define (determine-protocols)
  (walk-lambdas (lambda (l)
		  (cond ((and (eq? 'proc (lambda-type l))
			      (node? (node-parent l))
			      (not (calls-known? l))
			      (find-calls l))
			 => (lambda (calls)
			      (determine-lambda-protocol l calls)))))))

; PROC is called at the references in CALLS.  This collects the call and return
; protocols used in these calls and matches them with those used in PROC.

(define (determine-lambda-protocol node calls)
  (let ((call&cont-data (get-call&cont-data calls))
	(return-data (get-return-data node))
	(proc-data (lambda-protocol node)))
    (let ((call-match (match-proc&call-data proc-data call&cont-data))
	  (return-match (match-return&cont-data return-data call&cont-data)))
      (walk (lambda (ref)
              (let ((call (node-parent ref)))
                (unknown-call->known-call call call-match return-match)))
            calls)
      (walk (lambda (ref)
              (let ((call (node-parent ref)))
                (unknown-return->known-return call return-match)))
            (variable-refs (car (lambda-variables node))))
      (argument-protocol-adjustment node call-match)
      (set-calls-known?! node)
      (values))))

(define (get-call&cont-data refs)
  (map (lambda (ref)
	 (literal-value (call-arg (node-parent ref) 2)))
       refs))

(define (get-return-data node)
  (map (lambda (ref)
	 (literal-value (call-arg (node-parent ref) 1)))
       (variable-refs (car (lambda-variables node)))))

(define (unknown-call->known-call call call-match return-match)
  (call-protocol-adjustment call call-match return-match)
  (set-calls-known?! (call-arg call 0))
  (remove-call-arg call 2)
  (set-call-primop! call (get-primop primop/call)))

(define (unknown-return->known-return return return-match)
  (return-protocol-adjustment return return-match)
  (remove-call-arg return 1)
  (set-call-primop! return (get-primop primop/return)))

;-------------------------------------------------------------------------------
; Return a list of all the reference to lambda-node L's value that call it.
; If not all can be identified then #F is returned.

(define (find-calls l)
  (let* ((refs (cond ((bound-to-variable l)
		      => variable-refs)
		     ((procedure-node? l)
		      (list l))
		     (else
		      #f)))
	 (all-calls? (and refs (every? called-node? refs))))
    (if (and (calls-known? l) (not all-calls?))
	(bug "cannot find calls for known lambda ~S" l))
    refs))

