
; Creating nodes and CPS converting calls and blocks.
;-------------------------------------------------------------------------------

(define (attach-call-to-last last call rep)
  (cond ((lambda-node? last)
	 (attach-body last call))
	(else
	 (receive (var last)
	     (insert-continuation rep last)
	   (attach-body last call)))))

(define (insert-continuation rep call)
  (if rep
      (let* ((var (make-variable 'v rep))
	     (cont (make-lambda-node 'c 'cont (list var))))
	(attach-continuation cont rep call)
	(values var cont))
      (let ((cont (make-lambda-node 'c 'cont '())))
	(attach-continuation cont rep call)
	(values #f cont))))

(define (attach-continuation cont rep call)
  (case (primop-id (call-primop call))
    ((return unknown-return)
     (attach call 0 cont))
    ((let)
     (cond ((empty? (call-arg call 0))
	    (attach call 0 cont))
	   ((lambda-node? cont)
	    (let ((var (car (lambda-variables (call-arg call 0)))))
	      (attach call 1 cont)
	      (change-lambda-type cont 'jump)
	      (set-calls-known?! cont)))
	   (else
	    (attach call 1 cont))))
    (else
     (let ((cont (if (lambda-node? cont)
		     cont
		     (reference-cont->lambda-cont cont rep))))
       (attach call 0 cont)))))

(define (reference-cont->lambda-cont reference-node rep)
  (let* ((value-var (make-variable 'v rep))
         (cont (make-lambda-node 'c 'cont (list value-var)))
	 (call (make-simple-return-call reference-node 
					rep
					(make-reference-node value-var))))
    (attach-body cont call)
    cont))

(define (make-simple-return-call cont rep value)
  (if (and (reference-node? cont)
	   (eq? 'known (variable-flag (reference-variable cont))))
      (let ((call (make-call-node (get-primop primop/jump) 2 0)))
	(attach call 0 cont)
	(attach call 1 value)
	call)
      (let ((call (make-call-node (get-primop primop/unknown-return) 3 0)))
	(attach call 0 cont)
	(attach call 1 (make-literal-node (return-protocol-spec rep) #f))
	(attach call 2 value)
	call)))

; Returns a new first and last.

(define (value-node->call-node value rep first last)
  (let ((call (make-call-node (get-primop primop/let) 2 1))
	(last (if (or (not last) (lambda-node? last))
		  last
		  (receive (var binder)
		      (insert-continuation #f last)
		    binder))))
    (attach call 1 value)
    (if last (attach-body last call))
    (values (if first first call) call)))

; Returns a value and a new last.

(define (call-node->value-node rep last)
  (receive (var binder)
      (insert-continuation rep last)
    (values (if var
		(make-reference-node var)
		(make-undefined-literal))
	    binder)))

; The above two combined; returns a new value, first, and last.

(define (value-node->reference-node value rep first last)
  (receive (first last)
      (value-node->call-node value rep first last)
    (receive (value last)
	(call-node->value-node rep last)
      (values value first last))))

(define (ensure-last-is-not-call last rep)
  (if (and last (call-node? last))
      (receive (var last)
	  (insert-continuation rep last)
	last)
      last))

;                    CPS CONVERSION OF CALLS
;=============================================================================

; Turn a call into a node.  Just a front for CPS-ARGS.

(define (make-primop-call exp)
  (really-make-primop-call (call-exp-proc   exp)
			   (call-exp-exits  exp)
			   (call-exp-args   exp)
			   (call-exp-type   exp)
			   (call-exp-source exp)
			   #f))
	
(define (make-return-call exp)
  (really-make-primop-call (get-primop primop/unknown-return)
			   0
			   (cons (make-quote-exp (return-exp-protocol exp) #f)
				 (return-exp-args exp))
			   (return-exp-type exp)
			   #f
			   #t))

(define (really-make-primop-call primop exits args type source return?)
  (let* ((space? (or return?
		     (and (= exits 0)
			  (not (primop-trivial? primop)))))
	 (call (make-call-node primop
			       (+ (length args) (if space? 1 0))
			       (if (or return? (not space?)) exits 1))))
    (receive (first last)
	(cps-args call (+ exits (if space? 1 0)) (list-tail args exits))
      (set-call-source! call source)
      (cond (space?
	     (if last (attach-body last call))
	     (values #f type (or first call) call))
	    ((= exits 0)
	     (values call type first last))
	    (else
	     (convert-primop-exits call first last exits type args))))))
      
;(define (make-throw primop rep args)
;  (let* ((call (make-call-node primop (length args) '0))
;         (first (cps-args call 0 args)))
;    (values first (make-cont-loc '#f '0 rep 'any))))

(define (convert-primop-exits call first last exits rep args)
  (let* ((join (make-variable 'j rep))
         (join-call (make-call-node (get-primop primop/let) '2 '1))
         (proc (make-lambda-node 'b 'cont (list join))))
    (set-variable-flag! join 'known)
    (do ((args args (cdr args))
         (i 0 (+ i 1)))
        ((>= i exits))
      (convert-primop-exit (car args) i call join))
    (set-variable-flag! join #f)
    (attach-body proc call)
    (attach join-call 0 proc)
    (if last (attach-body last join-call))
    (values #f rep (or first join-call) join-call)))

(define (convert-primop-exit exp index call cont-var)
  (let ((vars (lambda-exp-vars exp))
	(body (lambda-exp-body exp)))
    (let ((node (make-lambda-node 'g 'cont vars)))
      (attach-body node (make-block-with-cont-var body cont-var))
      (attach call index node))))

(define (make-lambda-call exp)
  (let ((call (make-call-node (get-primop primop/let)
			      (+ 1 (length (let-exp-vals exp)))
			      1))
        (node (make-lambda-node 'l 'cont (let-exp-vars exp))))
    (set-call-source! call (let-exp-source exp))
    (receive (first last)
	(cps-args call 1 (let-exp-vals exp))
      (if last (attach-body last call))
      (attach call 0 node)
      (receive (value rep body-first last)
	  (x->node (let-exp-body exp))
	(values value
		rep
		(or first call)
		(cond (body-first
		       (attach-body node body-first)
		       last)
		      (else node)))))))

; Record to hold information about arguments to calls.

(define-record-type arg
  (index          ; The index of this argument in the call.
   rank           ; The estimated cost of executing this node at run time.
   value          ; What X->NODE returned for this argument.
   first
   last
   )
  ())

(define (make-arg exp index)
  (receive (value rep first last)
           (x->node exp)
    (receive (value first last)
	(cond ((not value)
	       (receive (value last)
		   (call-node->value-node rep last)
		 (values value first last)))
	      ((node-side-affects? value)
	       (value-node->reference-node value rep first last))
	      ((and last (call-node? last))
	       (receive (var last)
		   (insert-continuation rep last)
		 (values (if var (make-reference-node var) #f) first last)))
	      (else
	       (values value first last)))
      (arg-maker index (node-rank first) value first last))))

; The actual work of CPS conversion.  Takes a list of ARG records and 
; builds them into a call, returning the first node of the resulting tree.
; If an argument is a call a continuation is made for it.

(define (cps-args call start args)
  (let ((arguments (make-arg-nodes args start)))
    (let loop ((args arguments) (first #f) (last #f))
      (if (null? args)
	  (values first last)
	  (let ((arg (car args)))
	    (attach call (arg-index arg) (arg-value arg))
	    (if (and last (arg-first arg))
		(attach-body last (arg-first arg)))
	    (loop (cdr args)
		  (or first (arg-first arg))
		  (or (arg-last arg) last)))))))

; Convert the elements of EXP into nodes (if they aren't already) and put
; them into an ARG record.  Returns the list of ARG records sorted
; by ARG-RANK.

(define (make-arg-nodes exp start)
  (do ((i start (+ i '1))
       (args exp (cdr args))
       (vals '() (if (not (empty? (car args)))
                     (cons (make-arg (car args) i) vals)
                     vals)))
      ((null? args)
       (sort-list vals
                  (lambda (v1 v2)
		    (> (arg-rank v1) (arg-rank v2)))))))

; Complexity analysis used to order argument evaluation.  More complex
; arguments are to be evaluated first.  This just counts reference nodes.

(define (node-rank first)
  (if (not first)
      0
      (complexity-analyze-vector (call-args first))))

(define (complexity-analyze node)
  (cond ((empty? node)
         '0)
        ((reference-node? node)
         '1)
        ((lambda-node? node)
	 (if (not (empty? (lambda-body node)))
	     (complexity-analyze-vector (call-args (lambda-body node)))
	     0))
	((call-node? node)
	 (complexity-analyze-vector (call-args node)))
        (else
         '0)))

(define (complexity-analyze-vector vec)
  (do ((i '0 (+ i '1))
       (q '0 (+ q (complexity-analyze (vector-ref vec i)))))
      ((>= i (vector-length vec))
       q)))

(define (make-block-with-cont-var body cvar)
  (receive (value rep first last)
      (x->node body)
    (let ((cont (make-reference-node cvar)))
      (cond ((and value last)
	     (let ((return (make-simple-return-call cont rep value)))
	       (attach-call-to-last last return rep)
	       first))
	    (last
	     (attach-continuation cont rep last)
	     first)
	    (else
	     (make-simple-return-call cont rep value))))))

(define (make-block exps)
  (let loop ((exps exps) (value #f) (rep #f) (first #f) (last #f))
    (if (null? exps)
	(values value rep first last)
	;; do something with the value of the previous expression
	(receive (first last)
	    (cond ((not value)
		   (values first (ensure-last-is-not-call last rep)))
		  ((not (node-side-affects? value))
		   (erase value)
		   (values first last))
		  (else
		   (receive (value first last)
		       (value-node->reference-node value rep first last)
		     (erase value)
		     (values first last))))
	  (receive (xvalue xrep xfirst xlast)
	      (x->node (car exps))
	    (if (and last xfirst)
		(attach-call-to-last last xfirst rep))
	    (loop (cdr exps)
		  xvalue xrep (or first xfirst) (or xlast last)))))))

(define (node-side-affects? node)
  (let label ((node node))
    (cond ((not (call-node? node))
	   #f)
	  ((or (not (memq? (primop-side-effects (call-primop node))
			   '(#f read)))
	       (not (primop-trivial? (call-primop node))))
	   #t)
	  (else
	   (let loop ((i 0))
	     (cond ((>= i (call-arg-count node))  #f)
		   ((label (call-arg node i))     #t)
		   (else
		    (loop (+ i 1)))))))))
