
;----------------------------------------------------------------------------
;                           SPECIAL FORMS
;
;                   QUOTE CALL RETURN BLOCK LAMBDA
;                + LETREC (for type-checking reasons)
;
;----------------------------------------------------------------------------

(define-record-type quote-exp
  (value
   (type))
  ())

(define make-quote-exp quote-exp-maker)

(define-record-type call-exp
  (proc
   exits
   (type)
   args
   source)
  ())

(define make-call-exp call-exp-maker)

(define-record-type let-exp
  (vars
   vals
   (body)
   source)
  ())

(define make-let-exp let-exp-maker)

(define-record-type return-exp
  (protocol
   type
   args)
  ())

(define make-return-exp return-exp-maker)

(define-record-type block-exp
  (exps)
  ())

(define make-block-exp block-exp-maker)

(define-record-type lambda-exp
  (id
   (return-type)
   protocol
   vars
   (body)
   source)
  ())

(define make-lambda-exp lambda-exp-maker)
(define (make-continuation-exp vars body)
  (make-lambda-exp #f #f #f vars body #f))

(define-record-type letrec-exp
  (vars
   vals
   (body)
   source)
  ())

(define make-letrec-exp letrec-exp-maker)

(define-record-type external-value
  (type)
  ())

(define make-external-value external-value-maker)

; Returns:
; VALUE  - a node or #F
; REP    - representation
; FIRST  - first call node or #F
; LAST   - last call or lambda node, or #F
;
; There are several possibilities
;      VALUE               FIRST       LAST
; 1.  call-or-value-node  #f          #f
; 2.  call-or-value-node  call-node   lambda-node
; 3.  #f                  call-node   call-or-lambda-node
;
; In 2. the value will always be a trivial call.  None of the other calls are
; trivial.

(define (x->node exp)
  (cond ((variable? exp)
	 (values (make-reference-node exp) (variable-type exp) #f #f))
	((quote-exp? exp)
	 (values (make-literal-node (quote-exp-value exp)
				    (quote-exp-type exp))
		 (quote-exp-type exp) #f #f))
	((block-exp? exp)
	 (make-block (block-exp-exps exp)))
	((return-exp? exp)
	 (make-return-call exp))
	((call-exp? exp)
	 (make-primop-call exp))
	((let-exp? exp)
	 (make-lambda-call exp))
	((lambda-exp? exp)
	 (lambda-exp->node exp))
	((letrec-exp? exp)
	 (letrec-exp->node exp))
        (else
         (x->node (error '"unknown syntax~% ~S" exp)))))

;(define (dereference-cell-call var)
;  (let-nodes ((call (contents 0 (* var)
;			      '('cell #f)
;			      '((variable-type var) #f))))
;    call))

(define (x->value-node exp)
  (receive (value rep first last)
      (x->node exp)
    (if last
	(bug "->VALUE-NODE did not get a value node for ~S" exp)
	value)))

(define (lambda-exp->node exp)
  (let* ((cvar (make-variable 'c (lambda-exp-return-type exp)))
         (node (make-lambda-node (lambda-exp-id exp)
				 'proc
				 (cons cvar (copy-list (lambda-exp-vars exp))))))
    (if (not (lambda-exp-protocol exp))
	(breakpoint "no protocol"))
    (if (not (lambda-exp-return-type exp))
	(breakpoint "no return type"))
    (set-lambda-protocol! node (lambda-exp-protocol exp))
    (set-lambda-source! node (lambda-exp-source exp))
    (attach-body node (make-block-with-cont-var (lambda-exp-body exp) cvar))
    (values node type/unknown #f #f)))

(define (letrec-exp->node exp)
  (let ((vals (map x->value-node (letrec-exp-vals exp)))
	(vars (letrec-exp-vars exp))
	(cont (make-lambda-node 'c 'cont '())))
    (receive (body rep first last)
	(x->node (letrec-exp-body exp))
      (let-nodes ((top (letrec1 1 l1))
		  (l1 ((x #f) . vars) (letrec2 1 cont (* x) . vals)))
	(set-call-source! top (letrec-exp-source exp))
	(if first (attach-body cont first))
	(values body rep top (if last last cont))))))

; (CATCH id . body)
; (THROW primop rep id . args)

; Manipulating Cells

;(define (make-cell-call source size)
;  `(call ,source 
;	 ,(get-primop primop/allocate) 0 ,type/unknown
;	 (quote cell #f)
;	 (quote ,size #f)))

;(define (make-cell-push-call source size)
;  `(call ,source 
;	 ,(get-primop primop/push) 0 ,type/unknown
;	 (quote cell #f)
;	 (quote ,size #f)))

(define (dereference-cell source cell type use)
  (make-call-exp (get-primop primop/contents)
		 0
		 type
		 (list cell (make-quote-exp use #f) (make-quote-exp type #f))
		 source))

(define (set-cell source cell value type use)
  (make-call-exp (get-primop primop/set-contents)
		 0
		 #f
		 (list cell
		       (make-quote-exp use #f)
		       (make-quote-exp type #f)
		       value)
		 source))

(define (make-undefined-value)
  (make-quote-exp the-undefined-value #f))

;(define (make-letrec-set var value source)
;  (set-cell source var value (variable-type var) 'letrec))

(define (exp->s-exp exp)
  (cond ((variable? exp)
	 (format #f "~S_~S" (variable-name exp) (variable-id exp)))
	((quote-exp? exp)
	 (list 'quote (quote-exp-value exp)))
	((block-exp? exp)
	 (cons 'begin (map exp->s-exp (block-exp-exps exp))))
	((return-exp? exp)
	 (cons 'return (map exp->s-exp (return-exp-args exp))))
	((call-exp? exp)
	 `(,(primop-id (call-exp-proc exp))
	   ,(call-exp-exits exp)
	   . ,(map exp->s-exp (call-exp-args exp))))
	((let-exp? exp)
	 `(let ,(map list
		     (map exp->s-exp (let-exp-vars exp))
		     (map exp->s-exp (let-exp-vals exp)))
	    ,(exp->s-exp (let-exp-body exp))))
	((lambda-exp? exp)
	 `(lambda ,(map exp->s-exp (lambda-exp-vars exp))
	    ,(exp->s-exp (lambda-exp-body exp))))
	((letrec-exp? exp)
	 `(letrec ,(map list
			(map exp->s-exp (letrec-exp-vars exp))
			(map exp->s-exp (letrec-exp-vals exp)))
	    ,(exp->s-exp (letrec-exp-body exp))))
        (else
         (error '"unknown syntax~% ~S" exp))))

(define (exp-source exp)
  (cond ((call-exp? exp)
	 (call-exp-source exp))
	((let-exp? exp)
	 (let-exp-source exp))
	((letrec-exp? exp)
	 (letrec-exp-source exp))
	((lambda-exp? exp)
	 (lambda-exp-source exp))
	(else
	 #f)))

(define (find-some-source top-exp exp)
  (or (exp-source exp)
      (call-with-current-continuation
	(lambda (exit)
	  (let recur ((at top-exp))
	    (let ((hit? (cond ((eq? at exp)
			       #t)
			      ((call-exp? at)
			       (or (recur (call-exp-proc at))
				   (any recur (call-exp-args at))))
			      ((let-exp? at)
			       (or (recur (let-exp-body at))
				   (any recur (let-exp-vals at))))
			      ((letrec-exp? at)
			       (or (recur (letrec-exp-body at))
				   (any recur (letrec-exp-vals at))))
			      ((return-exp? at)
			       (any recur (return-exp-args at)))
			      ((lambda-exp? at)
			       (recur (lambda-exp-body at)))
			      ((block-exp? at)
			       (any recur (block-exp-exps at)))
			      (else #f))))
	      (if (and hit? (exp-source at))
		  (exit (exp-source at)))
	      hit?))))))
  