
(define-interface utilities-signature
  (export bug
	  user-error
	  true false
	  or-map
	  remove-similar-elts
	  select-from-table
	  table->list table->entry-list
	  table-push table-pop
	  merge-lists
	  vector-every?
	  make-ignorable
	  sub-vector->list
	  flag-assq

	  enforce
	  writec
	  mem?
	  walk walk-vector
	  vector-replace
	  copy-list
	  symbol-hash
	  string-hash
	  char->ascii
	  object-hash
	  union intersection set-difference

	  make-xvector xvector-length xvector-ref xvector-set! xvector->vector

	  strongly-connected-components

	  (define-subrecord :syntax)
	  (define-simple-record-type :syntax)
	  (define-local-syntax :syntax)
	  ))

(define-interface arch-signature
  (export primop
	  primop-count
	  primop/call     ; ugh.
	  primop/return
	  primop/jump
	  primop/throw
	  primop/unknown-call
	  primop/unknown-return
	  primop/dispatch
	  primop/let
	  primop/letrec1
	  primop/letrec2
	  primop/undefined-value
	  primop/undefined-effect
	  primop/allocate
	  primop/contents
	  primop/set-contents
	  primop/push
	  ))

(define-interface primop-signature
  (export primop? make-primop make-proc-primop make-conditional-primop
	  all-primops get-primop

	  primop-id primop-trivial? primop-side-effects
	  primop-cost
	  simplify-call
	  primop-procedure? primop-call-index
	  primop-conditional?
	  expand-to-conditional
	  simplify-conditional?
	  primop-code-data set-primop-code-data!
	  ))

(define-interface variable-signature
  (export variable? make-variable
	  global-variable? make-global-variable
	  variable-name set-variable-name!
	  variable-id
	  variable-type   set-variable-type!
	  variable-binder set-variable-binder!
	  variable-refs   set-variable-refs!
	  variable-flag   set-variable-flag!
	  variable-flags  set-variable-flags!
	  erase-variable
	  variable-index copy-variable used? unused?
	  known-value? variable-known-value add-known-value remove-known-value
	  ))

(define-interface node-signature
  (compound-interface
   arch-signature
   primop-signature
   variable-signature
   (export reset-node-id node-hash node-unhash

	   node? node-variant
	   node-parent set-node-parent!
	   node-index set-node-index!
	   node-simplified? set-node-simplified?!
	   node-flag set-node-flag!

	   empty empty? proclaim-empty

	   erase

	   detach detach-body
	   attach attach-body
	   move move-body
	   insert-body
	   replace replace-body

	   mark-changed

	   leaf-node?
	  
	   literal-node? make-literal-node
	   literal-value set-literal-value!
	   literal-type  set-literal-type!
	   copy-literal-node

	   reference-node? make-reference-node
	   reference-variable set-reference-variable!

	   call-node? make-call-node
	   call-primop set-call-primop!
	   call-args set-call-args!
	   call-exits set-call-exits!
	   call-source set-call-source!
	   call-arg call-arg-count

	   lambda-node? make-lambda-node
	   lambda-body set-lambda-body!
	   lambda-variables set-lambda-variables!
	   lambda-name set-lambda-name!
	   lambda-id
	   lambda-type set-lambda-type!
	   lambda-block set-lambda-block!
	   lambda-protocol set-lambda-protocol!
	   lambda-source set-lambda-source!
	   lambda-variable-count
	   calls-known? set-calls-known?!

	   initialize-lambdas add-lambda add-lambdas
	   change-lambda-type
	   walk-lambdas make-lambda-list
	  
	   loc/owner loc/type loc/rep
	   set/owner set/type set/rep set/value

	   node-base
	   trivial? nontrivial?
	   nontrivial-ancestor
	   calls-this-primop?
	   bound-to-variable
	   walk-refs-safely
	   call-exit?
	   small-node?
	   side-effects?
	   called-node?
	   called-lambda
	   get-lambda-value
	   ;set-reference?

	   attach-call-args remove-call-args replace-call-args
	   remove-null-arguments
	   shorten-call-args insert-call-arg remove-call-arg
	   append-call-arg

	   remove-body

	   remove-lambda-variable remove-variable remove-unused-variables

	   substitute substitute-vars-in-node-tree
	   replace-call-with-value

	   copy-node-tree

	   mark-ancestors marked-ancestor? unmarked-ancestor?
	   node-ancestor? marked-ancestor least-common-ancestor
	   proc-ancestor

	   hoistable-node?

	   find-scoping

	   (let-nodes :syntax)

	   node-equal?

           no-free-references?

	   pp-cps rereadable-pp-cps print-variable-name reset-pp-cps

	   node->vector vector->node vector->leaf-node

	   procedure-node? called-procedure-node

	   node-type

	   the-undefined-value
	   undefined-value?
	   undefined-value-node?
	   make-undefined-literal
	   )))

(define-interface simplify-signature
  (export simplify-node
	  default-simplifier
	  simplify-args
	  simplify-lambda-body
	  
	  (pattern-simplifier :syntax)

	  simplify-let
	  simplify-allocation
	  simplify-known-call
	  simplify-unknown-call
	  simplify-return
	  simplify-jump
	  ; simplify-undefined-value
	  simplify-test expand-test simplify-test?
	  ))

(define-interface front-debug-signature
  (export debug-breakpoint
	  add-checks add-check clear-checks clear-check
	  add-procs add-proc clear-procs clear-proc))

(define-interface front-signature
  (export make-quote-exp quote-exp? quote-exp-value quote-exp-type
	   set-quote-exp-type!
	  make-call-exp call-exp? call-exp-proc call-exp-exits call-exp-type
	   set-call-exp-type! call-exp-args call-exp-source
	  make-return-exp return-exp? return-exp-protocol return-exp-type
	   return-exp-args
	  make-block-exp block-exp? block-exp-exps
	  make-lambda-exp lambda-exp? lambda-exp-id lambda-exp-return-type
	   lambda-exp-protocol lambda-exp-vars lambda-exp-body
	   lambda-exp-source set-lambda-exp-return-type!
	   set-lambda-exp-body!
	   make-continuation-exp
	  make-let-exp let-exp? let-exp-vars let-exp-vals let-exp-body
	   let-exp-source
	   set-let-exp-body!
	  make-letrec-exp letrec-exp? letrec-exp-vars letrec-exp-vals
	   letrec-exp-body letrec-exp-source
	   set-letrec-exp-body!

	  make-undefined-value

	  make-external-value external-value? external-value-type

	  expand-and-simplify-expression
	  simplify-all
	  x->node integrate-jump-procs  ; for debugging

	  find-some-source

	  ;make-cell-call make-cell-push-call
	  dereference-cell set-cell
	  ))

(define-interface annotated-read-signature
  (export read-and-annotate
	  pair-annotation
	  annotated-cons
	  annotation?
	  annotation-file
	  annotation-form
	  annotation-row
	  annotation-column
	  ))

(define-interface byte-vector-signature
  (export make-byte-vector byte-vector? byte-vector-length
	  byte-vector-ref  byte-vector-word-ref  byte-vector-half-word-ref
	  byte-vector-set! byte-vector-word-set! byte-vector-half-word-set!
	  byte-vector-endianess set-byte-vector-endianess!
	  ))

(define-interface parameter-signature
  (export lookup-primop
	  
	  control-protocol-return-spec
	  return-protocol-spec
	  
	  type/unknown

	  lambda-node-type

	  true-value
	  false-value

	  match-proc&call-data
	  match-return&cont-data
	  match-return-data
	  call-protocol-adjustment
	  return-protocol-adjustment
	  argument-protocol-adjustment
	  ))

; A signature from the byte-code compiler

; Not currently used

(define-interface nodes-signature
  (export classify
	  make-node
	  make-similar-node
	  node?
	  node-operator
	  node-operator-id
	  node-form
	  node-ref
	  node-set!
	  node-predicate
	  name-node-binding

	  scan-body

	  lookup
	  bind
	  bind1

	  make-operator-table
	  operator-table-ref
	  operator-define!
	  get-operator

	  name?
	  generated?
	  generated-env
	  generated-symbol

	  schemify

	  n-ary?
	  normalize-formals

	  process-syntax
	  bind-evaluator-for-syntax
	  usual-transform))

(define-interface ps-primop-signature
  (export get-scheme-primop
	  make-primop-visible!
	  define-primitive-expander
	  (define-scheme-primop :syntax)
	  (define-nonsimple-scheme-primop :syntax)
	  (define-scheme-cond-primop :syntax)))

(define-interface ps-c-primop-signature
  (export simple-c-primop?
	  primop-generate-c
	  eval-primop-call
	  reconstruct-call-type
	  (define-c-generator :syntax)))

(define-interface evaluate-internal-signature
  (export eval-exp
	  eval-call
	  eval-call!
	  evaluate-list
	  env-set!
	  stored-object-ref
	  stored-object-set!
	  the-uninitialized-value
	  apply-closure
	  uninitialized-value?
	  ))

(define-interface ps-type-signature
  (export type/int7u
	  type/int8
	  type/int8u
	  type/int32
	  type/float32
	  type/float64

	  type/null
	  type/unit
	  type/boolean
	  type/undetermined
	  type/port

	  type/unknown
	  type/byte-pointer
	  type/char-pointer

	  other-type?
	  other-type-kind
	  other-type-subtypes
	  make-other-type
	  record-type?

	  base-type?
	  base-type-name

	  make-arrow-type
	  arrow-type?
	  arrow-type-result
	  arrow-type-args

	  make-pointer-type
	  pointer-type?
	  pointer-type-to

	  lookup-type

	  type-scheme?
	  schemify-type
	  instantiate-type-scheme
	  copy-type
	  type-scheme-type
	  make-nonpolymorphic!

	  type-scheme-free-uvars           ; for error messages
	  type-scheme-lattice-uvars
	  type-scheme-type

	  type-eq?
	  type>
	  type>=

	  lattice-type?

	  finalize-type

	  display-type
	  ))

(define-interface type-variable-signature
  (export make-uvar
	  uvar?
	  maybe-follow-uvar
	  set-uvar-source!

	  reset-type-vars!

	  uvar-binding set-uvar-binding!
	  uvar-supers
	  uvar-prefix
	  uvar-id
	  uvar-subs
	  uvar-arith-loc

	  bind-uvar!

	  unique-id

	  make-constrained-type
	  make-constrained-arith-monop-uvar
	  make-constrained-arith-binop-uvar

	  relation-source
	  relation-dest
	  relation-site
	  relation-temp set-relation-temp!

	  constrained-uvars
	  ))

(define-interface inference-signature
  (export type-check-exp
	  type-check-stob
	  get-variable-type
	  add-type-coercions))

(define-interface inference-internal-signature
  (export unify!
	  instantiate-type&value
	  make-monomorphic!
	  type-check
	  check-arg-type
	  check-call-arg-type
	  call-arg-types
	  type-check-cont
	  initialize-variable-type
	  reconstruct-variable-type
	  get-variable-type
	  literal-value-type
	  make-constrained-arith-monop-type
	  make-constrained-arith-binop-type
	  ))

(define-interface form-signature
  (export make-form
	  form?
	  form-value
	  set-form-value!
	  set-form-integrate!
	  set-form-exported?!
	  form-var
	  form-exported?
	  form-type
	  set-form-type!
	  form-free set-form-free!
	  set-form-node!
	  make-form-current
	  form-lambdas
	  form-name
	  variable-form
	  form-merge set-form-merge!
	  form-providers set-form-providers!
	  form-clients set-form-clients!
	    
	  erase-form
	  maybe-variable->form
	    
	  make-closure closure? closure-env
	  closure-exp set-closure-exp!
	  closure-free set-closure-free!
	  closure-variable set-closure-variable!
	  make-stob stob? stob-value
	  stob-type set-stob-type!
	  stob-variable set-stob-variable!

          ; high level stuff
	  sort-forms
	  expand-and-simplify-form
	  remove-unreferenced-forms
	  integrate-stob-form
	  resimplify-form
	  ))

(define-interface linking-interface
  (export package-specs->scanner+exports
	  define-pre-scheme!
	  environment-ref
	  environment-set!
	  
	  primitive?
	  primitive-compilator
	  primitive-variable
	  allow-references-to-primitives
	  ))

(define-interface c-internal-signature
  (export c-assignment
	  indent-to
	  c-argument-var
	  form-tail-called?
	  *doing-tail-called-procedure?*
	  merged-procedure-reference
	  goto-protocol?
	  save-renamed-variables
	  restore-renamed-variables
	  c-ify
	  *current-proc-var*
	  jump-to-procedure-start
	  c-value
	  form-value
	  form-name
	  variable-form
	  form-type
	  c-assign-to-variable
	  write-c-identifier
	  c-called-value
	  write-value-list
	  form-return-count
	  set-form-return-count!
	  simple-c-primop
	  c-variable
	  *current-merged-procedure*
	  write-c-block
	  write-c-coercion
	  no-value-node?
	  display-c-type
	  ))

