
; Utilities

(define-structures ((util utilities-signature))
  (open big-scheme scheme structure-refs)
  (for-syntax (open big-scheme scheme))
  (access primitives features)
  (files (util syntax) (util util) (util expand-vec) (util strong)))

(define-structures ((annotated-read annotated-read-signature))
  ; this is correct for linking, but doesn't work when loading
  ;(open defrecord extended-ports primitives scheme assembler)
  (open big-scheme scheme primitives fluids assembler)
  (files (prescheme track-read)))

(define-structures ((byte-vectors byte-vector-signature))
  (open scheme code-vectors bitwise signals)
  (files (util byte-vector)))

; Parameterizing the compiler

(define-structures ((parameters parameter-signature)
		    (set-parameters (export set-compiler-parameter!)))
  (open big-scheme scheme)
  (files param))

; PRIMitive OPerators

(define-structures ((arch arch-signature))
  (open scheme enumerated)
  (files (node arch)))

; The intermediate language (node tree)

(define-structures ((node     node-signature)
		    (variable variable-signature)
		    (primop   primop-signature))
  (open big-scheme scheme util arch parameters)
  (for-syntax (open big-scheme scheme let-nodes))
  (files (node syntax)
	 (node node) (node primop)
	 (node node-util) (node node-equal) (node pp-cps)
	 (node vector) (node leftovers)))

; Macro for creating node expressions

(define-structures ((let-nodes (export expand-let-nodes)))
  (open big-scheme scheme arch)
  (files (node let-nodes)))

; Matching procedures with call sites

(define-structures ((call (export find-calls
				  determine-protocols
				  determine-lambda-protocol)))
  (open big-scheme scheme util node parameters)
  (files (front call)))

; Converting tail-recursive calls to jumps

(define-structures ((jump (export integrate-jump-procs)))
  (open big-scheme scheme util node parameters call
	ps-types)  ; type-eq?  (should be a param probably)
  (files (front jump)))

; Replacing cells with values passed as parameters

(define-structures ((remove-cells (export remove-cells-from-tree)))
  (open scheme)
  (begin (define remove-cells-from-tree 'remove-cells-from-tree-is-undefined)))

; Flow analysis

(define-structures ((flow-values (export flow-values)))
  (open scheme)
  (begin (define flow-values 'flow-values-is-undefined)))

; Program simplification and partial evaluation

(define-structures ((simplify simplify-signature))
  (open big-scheme scheme util node parameters call)
  (for-syntax (open big-scheme scheme simp-patterns))
  (files (simp syntax)
	 (simp simplify) (simp let) (simp join)
	 (simp call) (simp cell) (simp global)))

; Macro for writing algebraic transformations

(define-structures ((simp-patterns (export make-pattern-simplifier)))
  (open big-scheme scheme)
  (files (simp pattern)))

; Translating the input program into a node program

(define-structures ((front       front-signature)
		    (front-debug front-debug-signature))
  (open big-scheme scheme util node simplify parameters call jump
	remove-cells flow-values)
  (files (front top) (front cps) (front special)))

; Pre-Scheme packages

; Translating Scheme into the compiler's input language

(define-structures ((expand (export expand-node expand schemify
				    variable-settable?)))
  (open big-scheme scheme structure-refs
	front variable arch primop scan util parameters
	syntactic meta-types ps-types linking)
  (access packages     ; package->environment, package-lookup
	  environments ; environment-ref
	  node)        ; the-undefined-value
  (for-syntax (open big-scheme scheme))
  (files (prescheme expand)))

; Processing signature and package definitions

(define-structures ((linking linking-interface))
  (open big-scheme scheme structure-refs
	scan interfaces packages environments usual-macros
	defpackage types ; to define pre-scheme-signature
	reflective-tower-maker	;Added this  -- JAR
	fluids
	syntactic        ; make-transform, get-operator, binding? binding-place
	locations)       ; contents
  (access meta-types           ; syntax-type
	  syntactic	       ; usual-variable-type
	  variable             ; make-global-variable
	  ps-types             ; type/unknown
	  packages-internal    ; $get-location
	  package-commands-internal ; config-package
	  built-in-structures) ; defpackage, structure-refs
  (files (prescheme ps-source)
	 (prescheme ps-signature)
	 (prescheme package)
	 (prescheme primitive)))

(define-structures ((ps-types ps-type-signature)
		    (type-variables type-variable-signature))
  (open big-scheme scheme util)
  (files (prescheme type)
	 (prescheme type-scheme)
	 (prescheme type-var)
	 ))

(define-structures ((inference inference-signature)
		    (inference-internal inference-internal-signature))
  (open big-scheme scheme front node util
	ps-types type-variables ps-primops ps-c-primops
	structure-refs
	forms)          ; for inserting coercions in forms and closures
  (access node        ; undefined-value?
	  syntactic)  ; schemify
  (for-syntax (open big-scheme scheme))
  (files (prescheme inference)
	 (prescheme infer-early)
	 (prescheme finalize)))

(define-structures ((ps-primops ps-primop-signature))
  (open big-scheme scheme util node simplify
	linking ps-types front expand)
  (files (prescheme/primop primop)))

(define-structures ((ps-c-primops ps-c-primop-signature))
  (open big-scheme scheme util node simplify
	ps-types ps-primops)
  (for-syntax (open big-scheme scheme))
  (files (prescheme/primop c-primop)))

(define-structures ((primop-data (export)))
  (open big-scheme scheme util node simplify call
	front expand type-variables inference-internal
	ps-types ps-primops)
  (files (prescheme/primop base)
	 (prescheme/primop arith)
	 (prescheme/primop io)
	 (prescheme/primop vector)))

(define-structures ((c-primop-data (export)))
  (open big-scheme scheme util node simplify
	ps-types ps-primops ps-c-primops
	front
	structure-refs
	c-internal
	ps-types type-variables inference-internal
	forms byte-vectors
	evaluate-internal)
  (access primitives)
  (files (prescheme/primop c-base)
	 (prescheme/primop c-arith)
	 (prescheme/primop c-io)
	 (prescheme/primop c-vector)))

; Evaluating top-level forms

(define-structures ((evaluate (export evaluate))
		    (evaluate-internal evaluate-internal-signature))
  (open big-scheme scheme front node byte-vectors
	ps-types ps-c-primops forms util)
  (files (prescheme eval)))

; Reducing closures and data structures to simple definitions

(define-structures ((flatten (export flatten))
		    (flatten-internal (export generated-top-variable?)))
  (open big-scheme scheme front variable
	evaluate ps-types forms util)
  (files (prescheme flatten)))

(define-structures ((forms form-signature))
  (open big-scheme scheme util node expand
	ps-primops ; get-scheme-primop
	simplify   ; simplify-node, simplify-args
	front      ; expand-and-simplify-expression
	ps-types   ; type/undetermined
	)
  (files (prescheme forms)))

(define-structures ((c (export write-c-file))
		    (c-internal c-internal-signature))
  (open scheme big-scheme util node forms
	ps-primops ps-c-primops
	ps-types type-variables
	flatten-internal   ; generated-top-variable?
	inference          ; get-variable-type
	inference-internal ; literal-value-type
	evaluate-internal  ; uninitialized-value?
	protocol           ; goto-protocol?
	front              ; quote-exp, quote-exp-type, quote-exp-value
	)
  (begin (define number-of-char-codes 256))  ; should be somewhere else
  (files (prescheme c)
	 (prescheme c-call)
	 (prescheme hoist)
	 (prescheme merge)))

(define-structures ((protocol (export goto-protocol?)))
  (open scheme big-scheme util set-parameters ps-primops ps-types node)
  (files (prescheme spec)))

; Everything else

(define-structures ((prescheme (export)))
  (open big-scheme scheme conditions defpackage util
	reflective-tower-maker  ;Don't know whether this is needed - JAR
	linking expand evaluate flatten
	set-parameters
	node front simplify
	call byte-vectors
	front-debug forms ps-primops
	ps-types
	type-variables
	inference
	inference-internal   ; instantiate, unify!
	c
	structure-refs
	primop-data
	c-primop-data
	)
  (for-syntax (open big-scheme scheme))
  (access primitives
	  syntactic  ; classify
	  packages)  ; package->environment
  (files (prescheme top)))

