
; Translating the node tree into C

(define (write-c-file init-name file forms shadow-data)
  (let ((out (make-tracking-output-port (open-output-file file))))
    (merge-forms forms)
    (format #t "Translating~%")
    (set! *c-variable-id* 0)
    (set! *type-uids* '())
    (set! *next-type-uid* 0)
    (set! *hoist-index* 0)
    (write-c-header out)
    (write-global-arg-variable-declarations forms out)
    (write-c-decls
     (filter-map (lambda (f)
                   (if (or (eq? (form-type f) 'stob)
			   (eq? (form-type f) 'initialize)
			   (eq? (form-type f) 'alias))
                       (form-var f)
                       '#f))
                 forms)
     out)
    (newline out)
    (for-each (lambda (f)
		(case (form-type f)
		  ((unused))
		  ((lambda)
		   (let ((vars (cond ((assq (variable-name (form-var f))
					    shadow-data)
				      => cdr)
				     (else
				      '()))))
		     (compile-proc-to-c f vars out)))
		  ((alias constant integrate merged stob initialize)
		   (values))
		  (else
		   (bug "unkown type of form ~S" f))))
	      forms)
    (write-c-main init-name out forms)
    (newline out)
    (set! *type-uids* '())
    (close-output-port out)))

(define (write-c-main init-name out forms)
  (set! *current-proc-var* #f)
  (set! *current-renamed-variables* #f)
  (set! *doing-tail-called-procedure?* #f)
  (set! *current-merged-procedure* #f)
  (cond ((any? (lambda (f)
		 (or (eq? (form-type f) 'initialize)
		     (eq? (form-type f) 'stob)
		     (eq? (form-type f) 'alias)))
	       forms)
	 (write-c-main-header (if init-name init-name 'main) out)
	 (for-each (lambda (f)
		     (case (form-type f)
		       ((initialize alias)
			(write-initialize (form-var f) (form-value f) out))
		       ((stob)
			(write-stob (form-var f)
				    (form-value f)
				    out))))
		   forms)
	 (write-c-main-end out))))

(define (write-c-header out)
  (format out "#include <stdio.h>~%")
  (format out "#include \"prescheme.h\"~%")
  (values))

(define (compile-proc-to-c form vars out)
  (format #t " ~A~%" (form-name form))
  (let* ((var (form-var form))
	 (name (c-ify (variable-name var)))
	 (name (if (generated-top-variable? var)
		   (string-append "H" name)
		   name)))
    (proc->c name form vars out #f)
    (for-each erase-form (form-merged form))
    (erase (detach-body (lambda-body (form-value form)))) ; need to keep the
    (set-form-node! form (form-value form) '())))         ; lambda around for help
                                                          ; with calls to it

(define (no-value-node? node)
  (or (undefined-value-node? node)
      (and (reference-node? node)
	   (eq? type/unit (get-variable-type (reference-variable node))))))

;------------------------------------------------------------
; Initialization procedure at the end of the file (often called `main').

; Header for initialization code

(define (write-c-main-header initname out)
  (write-c-identifier initname out)
  (format out "()~%{"))

; Write declarations for VARS

(define (write-c-decls vars port)
  (write-variable-declarations vars #f port '0))

; Write the end of the initialization code

(define (write-c-main-end out)
  (format out "~&}"))

(define (write-initialize var value out)
  (let ((wants (maybe-follow-uvar (variable-type var))))
    (receive (value has)
	(cond ((variable? value)
	       (values value (get-variable-type value)))
	      ((quote-exp? value)
	       (values (quote-exp-value value) (quote-exp-type value)))
	      (else
	       (error "unknown kind of initial value ~S" value)))
      (cond ((not (or (uninitialized-value? value)
		      (undefined-value? value)))
	     (newline out)
	     (write-c-identifier (variable-name var) out)
	     (format out " = ")
	     (if (not (type>= wants has))
		 (write-c-coercion wants out))
	     (cond ((input-port? value)
		    (display "stdin" out))
		   ((output-port? value)
		    (display "stdout" out))
		   ((variable? value)
		    (c-variable value out))
		   (else
		    (c-literal-value value has out)))
	     (writec out '#\;))))))

(define (write-stob var call out)
  (let* ((value (stob-value (literal-value (call-arg call 0))))
	 (type (stob-type value))
	 (wants (get-variable-type var)))
    (newline out)
    (write-c-identifier (variable-name var) out)
    (format out " = ")
    (cond ((vector? value)
	   (if (not (type-eq? type wants))
	       (write-c-coercion wants out))
	   (format out "malloc(~D * sizeof(" (vector-length value))
	   (display-c-type (pointer-type-to type) #f out)
	   (format out "));")
	   (do ((i 0 (+ i 1)))
	       ((>= i (vector-length value)))
	     (let* ((elt (call-arg call (+ i 1)))
		    (has (finalize-type
			  (if (reference-node? elt)
			      (variable-type (reference-variable elt))
			      (literal-value-type (literal-value elt))))))
	       (newline out)
	       (c-variable var out)
	       (format out "[~D] = " i)
	       (if (not (type>= (pointer-type-to type) has))
		   (write-c-coercion (pointer-type-to type) out))
	       (c-value elt out)
	       (write-char #\; out))))
	  (else
	   (error "don't know how to generate stob value ~S" value)))))

;------------------------------------------------------------
; Writing out a procedure.

(define (proc->c id form rename-vars port maybe-merged-count)
  (let ((top     (form-value form))
	(lambdas (form-lambdas form))
	(merged  (form-merged form))
	(tail?   (form-tail-called? form))
	(lambda-parent node-flag)	; Rename a couple of handy fields
	(lambda-kids lambda-block))
    (receive (procs others)
	(find-scoping lambdas
		      node-flag set-node-flag!
		      lambda-block set-lambda-block!)
      (for-each (lambda (l) (set-node-flag! l #f)) lambdas) ; clear NODE-FLAG
      (for-each (lambda (l)
		  (cond ((eq? l top))
			((no-free-references? l)
			 (hoist-lambda l)
			 (real-proc->c (lambda-name l) #f l
				       (lambda-kids l)
				       '() '() port #f))
			(else
			 (error "procedure ~S in ~S contains free references"
				(lambda-name l) id))))
		procs)
      (if maybe-merged-count
	  (merged-proc->c id top lambdas merged maybe-merged-count port tail?)
	  (real-proc->c id (form-var form) top (lambda-kids top)
			merged rename-vars port tail?))
      (values))))

(define (write-merged-form form port)
  (format #t "  ~A~%" (form-name form))
  (proc->c (form-name form)
	   form
	   '()
	   port
	   (length (variable-refs (form-var form)))))

;------------------------------------------------------------

; 1. write the header
; 2. declare the local variables and free identifiers
; 3. write out the body
; 4. write out all of the label lambdas

(define *current-proc-var* #f)
(define *current-renamed-variables* #f)
(define *doing-tail-called-procedure?* #f)
(define *current-merged-procedure* #f)

; These global variables should be replaced with fluids.

(define (real-proc->c id var top lambdas merged rename-var-ids port tail?)
  (let ((vars (cdr (lambda-variables top)))
	(return-type (get-variable-type (car (lambda-variables top))))
	(body (lambda-body top)))
    (set! *current-proc-var* var)
    (set! *current-renamed-variables* rename-var-ids)
    (set! *doing-tail-called-procedure?* tail?)
    (set! *current-merged-procedure* #f)
    (write-procedure-header id return-type vars port tail?)
    (write-char '#\{ port)
    (newline port)
    (let ((rename-vars (write-extern-declarations body rename-var-ids port)))
      (for-each (lambda (v)
		  (set-variable-flags! v (cons 'shadowed (variable-flags v))))
		rename-vars)
      (write-arg-variable-declarations lambdas merged port)      
      (write-rename-variable-declarations rename-vars port)
      (write-merged-declarations merged port)
      (format port (if tail? "~% START: {" "~% {"))
      (if tail? 
	  (write-global-argument-initializers (cdr (lambda-variables top))
					      port 2))
      (write-c-block body port 2)
      (for-each (lambda (f)
		  (write-merged-form f port))
		(reverse merged))  ; makes for more readable output
      (write-char '#\} port)
      (newline port)
      (for-each (lambda (v)
		  (set-variable-flags! v (delq! 'shadowed (variable-flags v))))
		rename-vars)
      (set! *current-proc-var* #f)            ; drop pointers
      (set! *current-renamed-variables* '())
      (values))))

(define (merged-proc->c id top lambdas merged return-count port tail?)
  (let ((vars (cdr (lambda-variables top)))
	(body (lambda-body top))
	(name (c-ify id)))
    (set! *doing-tail-called-procedure?* tail?)
    (set! *current-merged-procedure* name)
    (write-merged-header name top port)
    (write-extern-declarations body '() port)
    (write-arg-variable-declarations lambdas merged port)      
    (write-merged-declarations merged port)
    (write-char '#\{ port)
    (write-c-block body port 2)
    (if (not tail?)
	(write-merged-return name return-count port))
    (for-each (lambda (f)
		(write-merged-form f port))
	      (reverse merged))  ; makes for more readable output
    (write-char '#\} port)
    (newline port)
    (values)))

(define (write-merged-header name top port)
  (format port "~% ~A: {~%" name)
  (if (not (null? (cdr (lambda-variables top))))
      (write-merged-argument-initializers (cdr (lambda-variables top)) port 2)))

(define (write-merged-return name return-count port)
  (format port "~% ~A_return:~%  switch (~A_return_tag) {~%" name name)
  (do ((i 0 (+ i 1)))
      ((>= i return-count))
    (format port "  case ~S: goto ~A_return_~S;~%" i name i))
  (format port "  }"))

(define (write-merged-declarations forms port)
  (for-each (lambda (f)
	      (if (not (form-tail-called? f))
		  (write-merged-declaration f port)))
	    forms))

(define (write-merged-declaration form port)
  (let ((name (c-ify (form-name form)))
	(return-type (get-variable-type
		      (car (lambda-variables (form-value form))))))
    (format port "~%  int ~A_return_tag;" name)
    (cond ((not (eq? return-type type/unit))
	   (format port "~%  ")
	   (display-c-type-decl return-type
				(lambda (port)
				  (format port "~A_return_value" name))
				port)
	   (writec port #\;)))))

(define (write-procedure-header id return-type vars port tail?)
  (newline port)
  (display-c-type return-type
		  (lambda (port)
		    (if tail? (write-char #\T port))
		    (display id port))
		  port)
  (cond ((or tail? (null? vars))
	 (display "()" port))
	(else
	 (write-char '#\( port)
	 (write-variables vars port)
	 (write-char '#\) port)))
  (newline port))

; Write the names of VARS out to the port.

(define (write-variables vars port)
  (let ((do-one (lambda (var)
		  (display-c-type (get-variable-type var)
				  (lambda (port)
				    (c-variable var port))
				  port))))
    (cond ((null? vars)
	   (values))
	  ((null? (cdr vars))
	   (do-one (car vars)))
	  (else
	   (do-one (car vars))
	   (do ((vars (cdr vars) (cdr vars)))
	       ((null? vars)
		(values))
	     (write-char '#\, port)
	     (write-char '#\space port)
	     (do-one (car vars)))))))
  
; GATHER-FREE-VARS returns lists of free variables and called procedures.
; Declarations are written out for all of them.

(define (write-extern-declarations body renamed-variables port)
  (let ((extern (gather-free-vars body)))
    (cond ((not (null? extern))
           (write-variable-declarations extern #t port '2)
           (newline port)))
    (filter (lambda (var)
	      (memq? (variable-name var) renamed-variables))
	    extern)))

(define (write-rename-variable-declarations vars port)
  (for-each (lambda (var)
	      (indent-to port 2)
	      (display-c-type (get-variable-type var)
			      (lambda (port)
				(writec port #\R)
				(write-c-identifier (variable-name var) port))
			      port)
	      (display " = " port)
	      (write-c-identifier (variable-name var) port)
	      (format port ";~%"))
	    vars))

(define (save-renamed-variables port indent)
  (save-or-restore-renamed-variables #t port indent))

(define (restore-renamed-variables port indent)
  (save-or-restore-renamed-variables #f port indent))

(define (save-or-restore-renamed-variables save? port indent)
  (for-each (lambda (var)
	      (indent-to port indent)
	      (if (not save?) (write-char '#\R port))
	      (write-c-identifier var port)
	      (display " = " port)
	      (if save? (write-char '#\R port))
	      (write-c-identifier var port)
	      (format port ";~%"))
	    *current-renamed-variables*))

(define (write-c-block body port indent)
  (write-c-block-with-args body '() port indent))

(define (write-c-block-with-args body arg-vars port indent)
  (receive (body vars lambdas)
      (gather-local-vars body)
    (if (not (null? arg-vars))
	(write-argument-initializers arg-vars port indent))
    (cond ((not (null? vars))
	   (write-variable-declarations vars #f port indent)
	   (newline port)))
    (call->c body port indent)
    (for-each (lambda (l)
		(jump-lambda->c l port (- indent '2)))
	      lambdas)
    (write-char '#\} port)
    (values)))

(define (jump-lambda->c node port indent)
  (newline port)
  (indent-to port indent)
  (display " L" port)
  (display (lambda-id node) port)
  (display ": {" port)
  (newline port)
  (write-c-block-with-args (lambda-body node)
			   (lambda-variables node)
			   port
			   (+ '2 indent)))

(define (write-argument-initializers arg-vars port indent)
  (really-write-argument-initializers arg-vars "arg" port indent))

(define (write-merged-argument-initializers arg-vars port indent)
  (really-write-argument-initializers arg-vars "merged_arg" port indent))

(define (write-global-argument-initializers arg-vars port indent)
  (really-write-argument-initializers arg-vars "goto_arg" port indent))

(define (really-write-argument-initializers arg-vars name port indent)
  (do ((i 0 (+ i 1))
       (vars arg-vars (cdr vars)))
      ((null? vars) (values))
    (if (used? (car vars))
	(let* ((var (car vars))
	       (type (get-variable-type var)))
	  (cond ((not (eq? type/unit type))
		 (indent-to port indent)
		 (display-c-type type
				 (lambda (port) (c-variable var port))
				 port)
		 (display " = " port)
		 (display (c-argument-var name type i port) port)
		 (write-char '#\; port)))))))

(define (c-argument-var name type i port)
  (format #f "~A~DK~D" name (type->uid type) i))

(define *type-uids* '())
(define *next-type-uid* 0)

(define (type->uid type)
  (cond ((any (lambda (p)
		(type-eq? type (car p)))
	      *type-uids*)
	 => cdr)
	(else
	 (let ((id *next-type-uid*))
	   (set! *next-type-uid* (+ id 1))
	   (set! *type-uids* (cons (cons type id) *type-uids*))
	   id))))

;------------------------------------------------------------
; Writing declarations.

(define declaration-column-limit 70)

(define (write-variable-declarations vars extern? port indent)
  (for-each (lambda (var)
	      (if (not (eq? type/unit (get-variable-type var)))
		  (really-write-variable-declaration var extern? port indent)))
	    vars))

(define (really-write-variable-declaration var extern? port indent)
  (indent-to port indent)
  (if extern? (display "extern " port))
  (display-c-type-decl (get-variable-type var)
		       (lambda (port)
			 (if (memq? 'tail-called (variable-flags var))
			     (writec port #\T))
			 (c-variable-no-shadowing var port))
		       port)
  (writec port #\;))

(define (write-c-coercion type out)
  (write-char #\( out)
  (display-c-type type #f out)
  (write-char #\) out))
  
(define (display-c-type type name port)
  (display-c-base-type (type->c-base-type type) port)
  (if name (display " " port))
  (display-c-type-modifiers type name port #f))

(define (display-c-type-decl type name port)
  (display-c-base-type (type->c-base-type type) port)
  (if name (display " " port))
  (display-c-type-modifiers type name port #t))

(define (display-c-type-modifiers type name port decl?)
  (let label ((type type) (name name))
    (let ((type (maybe-follow-uvar type)))
      (cond ((base-type? type)
	     (if name (name port)))
	    ((pointer-type? type)
	     (label (pointer-type-to type)
		    (lambda (port)
		      (format port "*")
		      (if name (name port)))))
	    ((arrow-type? type)              ; Argh!  Depending on context,
	     (if (not decl?)                 ; + is either int +(int,int)
		 (format port "(*"))         ; or int (*+)(int,int)
	     (label (arrow-type-result type) name)
	     (format port (if decl? "(" ")("))
	     (let ((args (arrow-type-args type)))
	       (cond ((not (null? args))
		      (display-c-type (car args) #f port)
		      (do ((args (cdr args) (cdr args)))
			  ((null? args))
			(display ", " port)
			(display-c-type (car args) #f port))))
	       (format port ")")))
	    (else
	     (bug "don't know how to write ~S as a C type" type))))))

(define (display-c-base-type type port)
  (display (case (base-type-name type)
	     ((boolean) "unsigned char")
	     ((int7u) "char")  ; hmm.  this may be a problem...
	     ((int8)  "char")
	     ((int8u) "unsigned char")
	     ((int16 int8+eof) "int")
	     ((int32) "long")
	     ((port) "FILE *")
	     ((unit) "void")
	     ((null) "void")
	     (else (bug "not a base type ~S" type)))
	   port))

(define (type->c-base-type type)
  (let ((type (maybe-follow-uvar type)))
    (cond ((base-type? type) type)
	  ((pointer-type? type)
	   (type->c-base-type (pointer-type-to type)))
	  ((arrow-type? type)
	   (type->c-base-type (arrow-type-result type)))
	  (else
	   (bug "don't know how to write ~S as a C type" type)))))

(define (old-really-write-variable-declarations decl vars port indent procedures?)
  (let* ((increment (if procedures? 4 2))
	 (do-it (lambda (var port space first?)
		  (let ((space (- space (+ (variable-name-space var)
					   increment))))
		    (cond ((or first? (> space 0))
			   (if (not first?)
			       (format port ", "))
			   (c-variable-no-shadowing var port)
			   (if procedures? (display "()" port))))
		    space)))
	 (space (- declaration-column-limit (+ indent (string-length decl)))))
    (let loop ((vars vars))
      (cond ((not (null? vars))
             (indent-to port indent)
             (display decl port)
             (writec port #\space)
	     (let ((space (do-it (car vars) port space #t)))
	       (if (null? (cdr vars))
		   (writec port #\;)
		   (loop (write-some-var-decls (cdr vars) space do-it port)))))))))

(define (write-some-var-decls vars space do-it port)
  (let loop ((vars vars) (space space))
    (let ((space (do-it (car vars) port space #f)))
      (cond ((<= space 0)
	     (writec port #\;)
	     vars)
	    ((null? (cdr vars))
	     (writec port #\;)
	     '())
	    (else
	     (loop (cdr vars) space))))))

(define (variable-name-space var)
  (string-length (if (string? var)
		     var
		     (symbol->string (variable-name var)))))

;------------------------------------------------------------

(define (gather-local-vars node)
  (let ((parent (node-parent node)))
    (receive (vars lambdas)
	(really-gather-local-vars node)  ; may change (lambda-body parent)
      (values (lambda-body parent) vars lambdas))))

(define (really-gather-local-vars node)
  (let loop ((call node) (vars '()) (lambdas '()))
    (let* ((call (expand-nasty-c-primops! call))
	   (primop (call-primop call))
	   (args   (call-args   call))
	   (exits  (call-exits  call)))
      (cond ((eq? (primop-id primop) 'let)
	     (let ((proc (vector-ref args 0)))
	       (receive (new-vars new-lambdas)
		   (gather-local-vars-in-let proc args)
		 (loop (lambda-body proc)
		       (append new-vars vars)
		       (append new-lambdas lambdas)))))
	    ((or (not (= exits 1))
		 (goto-call? call))
	     (values (reverse! vars) lambdas))
	    (else
	     (let ((cont (vector-ref args 0)))
	       (cond ;((puts-lambda-in-cell primop call)
		     ; => (lambda (l)
			;   (loop (lambda-body cont) vars (cons l lambdas))))
		     ((eq? (primop-id primop) 'letrec2)
		      (loop (lambda-body cont)
			    vars
			    (append (sub-vector->list (call-args call) 2)
				    lambdas)))
		     ((eq? (primop-id primop) 'letrec1)
		      (loop (lambda-body cont) vars lambdas))
		     (else
		      (loop (lambda-body cont)
			    (append (filter used? (lambda-variables cont)) vars)
			    lambdas)))))))))

(define (gather-local-vars-in-let proc args)
  (let loop2 ((i 1) (vs (lambda-variables proc)) (vars '()) (lambdas '()))
    (cond ((>= i (vector-length args))
	   (values vars lambdas))
	  ((lambda-node? (vector-ref args i))
	   (loop2 (+ i 1) (cdr vs)
		  vars (cons (vector-ref args i) lambdas)))
	  (else
	   (loop2 (+ i 1) (cdr vs)
		  (if (used? (car vs))
		      (cons (car vs) vars)
		      vars)
		  lambdas)))))

(define (goto-call? call)
  (and (eq? 'unknown-call (primop-id (call-primop call)))
       (goto-protocol? (literal-value (call-arg call 2)))))

;(define (puts-lambda-in-cell primop call)
;  (if (and (eq? 'set-contents (primop-id primop))
;	   (eq? 'letrec (literal-value (call-arg call set/type))))
;      (call-arg call set/value)
;      #f))

(define (jumped-to? var)
  (any? (lambda (n)
	  (and (calls-this-primop? (node-parent n) 'jump)
	       (= 0 (node-index n))))
	(variable-refs var)))

; Some primops must be given continuations so that calls to them will
; be translated into separate C statements and so expand into arbitrarily
; complex chunks of C if necessary.

(define (expand-nasty-c-primops! call)
  (let ((top call))
    (let label ((call call))
      (cond ((call-node? call)
	     (if (and (= 0 (call-exits call))
		      (nasty-c-primop-call? call))
		 (set! top (expand-nasty-c-primop! call top)))
	     (walk-vector label (call-args call)))))
    top))

(define (nasty-c-primop-call? call)
  (case (primop-id (call-primop call))
    ((read-char) #t)    ; may need to do `clearerr' after reading an EOF
    ((ashl ashr)        ; C does poorly when shifting by large amounts
     (not (literal-node? (call-arg call 1))))
    (else #f)))

; Give CALL a continuation and move it above TOP, replacing CALL
; with the continuation's variable.
;
; top = (p1 ... (p2 a1 ...) ...)
;  =>
; (p2 (lambda (v) (p1 ... v ...)) a1 ...)

(define (expand-nasty-c-primop! call top)
  (let* ((var (make-variable 'x (node-type call)))
	 (cont (make-lambda-node 'c 'cont (list var))))
    (move call
	  (lambda (call)
	    (make-reference-node var)))
    (insert-body call
		 cont
		 (node-parent top))
    (set-call-exits! call 1)
    (insert-call-arg call 0 cont)
    call))

;------------------------------------------------------------
; Declare the variables used to pass arguments to procedures.
; This is done in each procedure so that the C compiler doesn't have to contend
; with the possibility of globally visible side-effects.

(define (write-arg-variable-declarations lambdas merged port)
  (let ((lambdas (filter (lambda (l)
			   (eq? 'jump (lambda-type l)))
			 lambdas))
	(merged (map form-value merged)))
    (really-write-arg-variable-declarations lambdas "arg" port 2)
    (really-write-arg-variable-declarations merged "merged_arg" port 2)))

(define (write-global-arg-variable-declarations forms port)
  (let ((lambdas (filter-map (lambda (f)
			       (if (and (form-var f)
					(memq? 'tail-called
					       (variable-flags (form-var f))))
				   (form-value f)
				   #f))
			     forms)))
    (really-write-arg-variable-declarations lambdas "goto_arg" port 0)))
  
(define (really-write-arg-variable-declarations lambdas name port indent)
  (for-each (lambda (data)
	      (destructure (((uid type . indicies) data))
		(if (not (eq? type type/unit))
		    (for-each (lambda (i)
				(indent-to port indent)
				(declare-arg-variable type uid i name port))
			      indicies))))
	    (get-variable-decl-data lambdas)))

(define (get-variable-decl-data lambdas)
  (let ((data '()))
    (for-each (lambda (l)
		(do ((vars (if (eq? 'jump (lambda-type l))
			       (lambda-variables l)
			       (cdr (lambda-variables l)))
			   (cdr vars))
		     (i 0 (+ i 1)))
		    ((null? vars))
		  (let* ((type (get-variable-type (car vars)))
			 (uid (type->uid type))
			 (datum (assq uid data)))
		    (cond ((not datum)
			   (set! data (cons (list uid type i) data)))
			  ((not (memq i (cddr datum)))
			   (set-cdr! (cdr datum) (cons i (cddr datum))))))))
	      lambdas)
    data))

(define (declare-arg-variable type uid i name port)
  (display-c-type type
		  (lambda (port)
		    (format port "~A~DK~D" name uid i))
		  port)
  (format port ";~%"))

(define (gather-free-vars node)
  (let ((vars '()))
    (define (do-call call)
      (let ((args (call-args call)))
	(do ((i '0 (+ i '1)))
	    ((>= i (vector-length args)))
	  (do-node (vector-ref args i)))))
    (define (do-node n)
      (cond ((lambda-node? n)
	     (do-call (lambda-body n)))
	    ((call-node? n)
	     (do-call n))
	    ((reference-node? n)
	     (let ((var (reference-variable n)))
	       (if (and (global-variable? var)
			(not (merged-procedure-reference n))
			(not (memq? var vars)))
		   (set! vars (cons var vars)))))))
    (do-node node)
    vars))

(define (used-as-label? node)
  (and (goto-call? (node-parent node))
       (eq? 1 (node-index node))))

; random type stuff

(define (reference-type node)
  (finalize-variable-type (reference-variable node)))

(define (finalize-variable-type var)
  (let ((type (finalize-type (variable-type var))))
    (if (uvar? type)
	(error "no type inferable for ~S" var))
    (set-variable-type! var type)
    type))

