
; Type checking.

; Entry point

(define (type-check-exp exp loc name)
  (set! *currently-checking* name)
  (set! *current-top-exp* exp)
  (let ((res (type-check exp 0 loc)))
    (set! *currently-checking* #f)
    (set! *current-top-exp* #f)
    res))

; Another entry point

(define (type-check-stob stob name)
  (set! *currently-checking* name)
  (let ((value (stob-value stob)))
    (cond ((vector? value)
	   (let ((uvar (make-uvar (if (stob-variable stob)
				      (variable-name (stob-variable stob))
				      'stob)
				  -1)))
	     (do ((i 0 (+ i 1)))
		 ((>= i (vector-length value)))
	       (receive (type new-value)
		   (type-check-thing (vector-ref value i))
		 (if new-value (vector-set! value i new-value))
		 (unify! uvar type stob)))
	     (let ((type (make-pointer-type (maybe-follow-uvar uvar))))
	       (set-stob-type! stob type)
	       (set! *currently-checking* #f)
	       type)))
	  (else
	   (error "unknown kind of stob ~S" value)))))

(define (type-check-thing thing)
  (if (variable? thing)
      (receive (type ignore)
	  (reconstruct-variable-type thing -1)
	(values type #f))
      (let ((type (literal-value-type thing)))
	(if (lattice-type? type)
	    (let ((exp (make-quote-exp thing type)))
	      (values (make-constrained-type 'stob type -1 exp)
		      exp))
	    (values type #f)))))

; Dispatch on the type of an expression.  If a call returns a coercable type,
; then a type variable is introduced to make coercions possible.

(define type-checker-table (make-table))

(define (type-check exp depth location)
  (receive (type coercion-name)
      (cond ((variable? exp)
	     (reconstruct-variable-type exp depth))
	    ((quote-exp? exp)
	     (let ((type (literal-value-type (quote-exp-value exp))))
	       (set-quote-exp-type! exp type)
	       (values type
		       (if (lattice-type? type)
			   (let* ((value (quote-exp-value exp))
				  (value (if (char? value)
					     (char->ascii value)
					     value)))
			     (concatenate-symbol 'q value))
			   #f))))
	    ((block-exp? exp)
	     (type-check-block (block-exp-exps exp) depth))
	    ((call-exp? exp)   (type-check-call exp depth))
	    ((let-exp? exp)    (type-check-let exp depth))
	    ((lambda-exp? exp) (type-check-lambda exp depth))
	    ((letrec-exp? exp) (type-check-letrec exp depth))
	    ((return-exp? exp)
	     (let ((args (return-exp-args exp)))
	       (values (make-return-type (type-check (car args) depth args))
		       #f)))
	    (else
	     (error "uncompilable object~%  ~S" exp)))
    (let ((type (maybe-follow-uvar type)))  ; is this really necessary?
      (if coercion-name
	  (make-constrained-type coercion-name type depth location)
	  (maybe-follow-uvar type)))))

(define-local-syntax (define-type-checker pattern vars . body)
  (let ((name (car pattern)))
    `(table-set! type-checker-table ',name
		 (lambda ,vars
		   (destructure ((,(cdr pattern) (cdr ,(car vars))))
		     ,@body)))))

; Get the type of the variable - if it is a type-variable, then create a new
; one and relate the two; if it is a polymorphic pattern, instantiate it.
; How to pass the source?

(define (reconstruct-variable-type var depth)
  (let ((type (get-variable-type var)))
    (values (if (type-scheme? type)
		(instantiate-type-scheme type depth)
		type)
	    (variable-name var))))

(define (get-variable-type var)
  (let ((rep (variable-type var)))
     (cond ((not (eq? rep type/undetermined))
	    rep)
	   ((global-variable? var)
	    (reconstruct-global-variable-type var))
	   (else
	    (error "lexically bound variable ~S has no type" var)))))

; If no type is present, create a type variable.

(define (reconstruct-global-variable-type var)
  (let ((has (maybe-follow-uvar (variable-type var))))
    (cond ((not (eq? has type/undetermined))
	   has)
	  (else
	   (let ((type (make-uvar (variable-name var) -1)))
	     (set-variable-type! var type)
	     (set-uvar-source! type var)
	     type)))))

(define (literal-value-type value)
  (cond ((boolean? value)
	 type/boolean)
	((char? value)
	 type/int8u)
	((integer? value)
	 (if (< value 0)
	     (if (< value -256)
		 type/int32
		 type/int8)
	     (cond ((< value 128) type/int7u)
		   ((< value 256) type/int8u)
		   (else type/int32))))
	((string? value)
	 type/char-pointer)
	(((structure-ref node undefined-value?) value)
	 type/unit)
	((or (input-port? value)
	     (output-port? value))
	 type/port)
	((external-value? value)
	 (external-value-type value))
	(else
	 (error "don't know type of literal ~S" value))))

(define (type-check-lambda exp depth)
  (let* ((uid (unique-id))
	 (var-types (map (lambda (var)
			   (if var
			       (initialize-variable-type var uid depth)
			       (make-uvar #f depth uid)))
			 (lambda-exp-vars exp)))
	 (result (type-check (lambda-exp-body exp) depth exp)))
    (if (not (return-type? result))
	(error "body of LAMBDA does not return ~S" exp))
    ; stash the return type
    (set-lambda-exp-return-type! exp (return-type-type result))
    (values (make-arrow-type var-types (return-type-type result))
	    #f)))
	
; Create a new type variable for VAR.

(define (initialize-variable-type var uid depth)
  (let ((tvar (make-uvar (variable-name var) depth uid)))
    (set-variable-type! var tvar)
    (set-uvar-source! tvar var)
    tvar))

(define (type-check-block exps depth)
  (let loop ((exps exps) (type type/unit))
    (if (null? exps)
	(values type #f)
	(loop (cdr exps)
	      (type-check (car exps) depth exps)))))

(define (type-check-call exp depth)
  (let* ((proc (call-exp-proc exp))
	 (type (maybe-follow-uvar (reconstruct-call-type proc exp depth))))
    (set-call-exp-type! exp type)
    (values type
	    (if (or (uvar? type)
		    (lattice-type? type))
		(primop-id proc)
		#f))))

(define (type-check-let exp depth)
  (let ((depth (+ depth 1))
	(uid (unique-id)))
    (do ((vars (let-exp-vars exp) (cdr vars))
	 (vals (let-exp-vals exp) (cdr vals)))
	((null? vars))
      (let ((type (schemify-type (type-check (car vals) depth vals) depth)))
	(if (type-scheme? type)
	    (set-variable-type! (car vars) type)
	    (unify! (initialize-variable-type (car vars) uid depth)
		    type
		    exp))))
    (values (type-check (let-exp-body exp) depth exp) #f)))

(define (type-check-letrec exp depth)
  (let ((depth (+ depth 1))
	(uid (unique-id))
	(vars (letrec-exp-vars exp))
	(vals (letrec-exp-vals exp)))
    (for-each (lambda (var)
		(initialize-variable-type var uid depth))
	      vars)
    (do ((vars vars (cdr vars))
	 (vals vals (cdr vals)))
	((null? vars))
      (unify! (type-check (car vals) depth vals)
	      (get-variable-type (car vars))
	      exp))
    (for-each (lambda (var)
		(let ((type (schemify-type (get-variable-type var) depth)))
		  (if (type-scheme? type)
		      (set-variable-type! var type))))
	      vars)
    (values (type-check (letrec-exp-body exp) depth exp)
	    #f)))

;--------------------------------------------------
; Return types (can't handle multiple value returns yet)

(define-record-type return-type
  (
   type     ; type returned
   )
  ())

(define make-return-type return-type-maker)

(define-record-discloser type/return-type
  (lambda (return)
    (list 'return-type
	  (return-type-type return))))

;--------------------------------------------------
; Utility procedures used by the inferencers of the various primops.

; Return a list of the types of ARGS.

(define (call-arg-types args depth)
  (do ((args args (cdr args))
       (res  '()  (cons (type-check (car args) depth args) res)))
      ((null? args)
       (reverse res))))

; CONT is a primop continuation that will be passed a value of type TYPE.

(define (reconstruct-cont-type cont type depth)
  (let ((vars (lambda-exp-vars cont))
	(body (lambda-exp-body cont)))
    (cond ((null? vars)
	   (if (neq? type type/null)
	       (type-conflict "continuation takes no arguments"
			      cont)))
	  ((null? (cdr vars))
	   (set-variable-type! (car vars) type))
	  (else
	   (type-conflict "continuation expects more than one argument"
			  cont)))
    (type-check body depth cont)))

; Check that the INDEX'th argument of CALL has type TYPE.

(define (check-arg-type args index type depth exp)
  (unify! (type-check (list-ref args index)
		      depth
		      (list-tail args index))
	  type
	  exp))

(define (check-call-arg-type call index type depth)
  (check-arg-type (call-exp-args call) index type depth call))

(define (type-check-cont cont depth)
  (let ((vars (lambda-exp-vars cont))
	(body (lambda-exp-body cont)))
    (type-check body depth cont)))
  
(define (make-constrained-arith-monop-type prefix type loc depth site)
  (let ((uvar (make-constrained-arith-binop-uvar prefix type loc depth site)))
    (if (not uvar)
	(arith-type-error site))
    uvar))
	
(define (make-constrained-arith-binop-type prefix type1 loc1 type2 loc2 depth site)
  (let ((uvar (make-constrained-arith-binop-uvar prefix type1 loc1 type2 loc2 depth site)))
    (if (not uvar)
	(arith-type-error site))
    uvar))
	
