
(define defpackage     (structure-ref built-in-structures defpackage))
(define structure-refs (structure-ref built-in-structures structure-refs))
(define scheme         (structure-ref built-in-structures scheme))

(define (make-env-for-syntax-promise scheme)
  (make-reflective-tower eval (list scheme) 'pre-scheme-linking))

(define (make-very-simple-package name opens)
  (make-simple-package opens
		       eval
		       (make-env-for-syntax-promise scheme)
		       name))

(define pre-scheme-primitives-package
  (make-very-simple-package 'pre-scheme-primitives '()))

(define pre-scheme-primitives
  (make-structure pre-scheme-primitives-package
		  (lambda () pre-scheme-primitives-interface)
		  'pre-scheme-primitives))

(define pre-scheme-internal-primitives
  (make-structure pre-scheme-primitives-package
		  (lambda () pre-scheme-internal-primitives-interface)
		  'pre-scheme-internal-primitives))

(for-each (lambda (name)
	    (package-define! pre-scheme-primitives-package
			     name
			     (make-transform
			      (usual-transform name)
			      pre-scheme-primitives-package
			      (structure-ref meta-types syntax-type)
			      `(usual-transform ',name)
			      name)))
	  '(and cond do let let* or quasiquote)) ; delay syntax-rules case

(define (define-pre-scheme! name nargs value)
  (package-define! pre-scheme-primitives-package
		   name
		   (structure-ref syntactic usual-variable-type)
		   (make-primitive name nargs value)))

(let ((syntax-type (structure-ref meta-types syntax-type)))
  (for-each (lambda (id)
	      (package-define! pre-scheme-primitives-package
			       id
			       (get-operator id syntax-type)))
	    '(if begin lambda letrec quote set!
		 define define-syntax let-syntax
		 goto external)))  ; special for Pre-Scheme

; Newly made for each compilation so that we get fresh variables.

(define (make-pre-scheme)
  (let ((package
	 (let-fluid (structure-ref packages-internal $get-location) get-variable
	   (lambda ()
	     (make-package (lambda () (list pre-scheme-primitives
					    pre-scheme-internal-primitives))
			   (lambda () '())
			   eval
			   (make-env-for-syntax-promise scheme)
			   ""
			   (list (cons 'begin pre-scheme-source))
			   #f
			   'pre-scheme)))))
    (make-structure package
		    (lambda () pre-scheme-interface)
		    'pre-scheme)))

(define (get-variable package name)
  ((structure-ref variable make-global-variable)
   name
   (structure-ref ps-types type/unknown)))

(define (get-reflective-tower-maker p)
  (environment-ref p (string->symbol ".make-reflective-tower.")))

(define (package-specs->scanner+exports name files)
  (let ((config (make-very-simple-package 'config (list defpackage)))
	(old-config ((structure-ref package-commands-internal config-package)))
	(pre-scheme (make-pre-scheme)))
    (environment-define! config 'pre-scheme            pre-scheme)
    (environment-define! config 'pre-scheme-primitives pre-scheme-primitives)
    (environment-define! config 'structure-refs        structure-refs)
    (set-reflective-tower-maker! config (get-reflective-tower-maker old-config))
    (let-fluids
	(structure-ref packages-internal $get-location)
	get-variable
	$note-file-package
	(lambda (filename package) (values))
	;; $for-syntax-evaluator (lambda (form) (eval form old-config))
      (lambda () 
	(for-each (lambda (file)
		    (load file config))
		  files)))
    (let ((struct (environment-ref config name)))
      (values (make-scanner-maker struct)
	      (structure-package pre-scheme)
	      (let ((names '()))
		(for-each-declaration (lambda (name type)
					(set! names (cons name names)))
				      (structure-interface struct))
		names)
	      (make-lookup config)))))

(define (make-lookup config)
  (lambda (package-id id)
    (let ((binding (package-lookup config package-id)))
      (if (and (binding? binding)
	       (location? (binding-place binding))
	       (structure? (contents (binding-place binding))))
	  (let* ((package (structure-package
			   (contents (binding-place binding))))
		 (binding (package-lookup package id)))
	    (if (binding? binding)
		(binding-place binding)
		#f))
	  #f))))
	       
(define (make-scanner-maker struct)
  (lambda (package-action)
    (let ((loaded '()))
      (let-fluid $note-file-package (lambda (filename package) (values))
	(lambda ()
	  (scan-structures
	   (list struct)
	   (lambda (package)
	     (not (memq package loaded)))
	   (lambda (stuff package)
	     ;; stuff = pair (file . (node1 node2 ...))
	     (for-each (lambda (filename+scanned-forms)
			 (package-action package
					 (cdr filename+scanned-forms)))
		       stuff)
	     (set! loaded (cons package loaded)))))))))
