
(define (reconstruct-arith-binop call depth)
  (let ((id (primop-id (call-exp-proc call)))
	(args (call-exp-args call)))
    (if (not (= 0 (call-exp-exits call)))
	(bug "who's making calls with real exits? ~S" call))
    (let ((type1 (type-check (car args) depth #f))
	  (type2 (type-check (cadr args) depth #f)))
      (make-constrained-arith-binop-type id
					 type1 args
					 type2 (cdr args)
					 depth call))))

(define (reconstruct-arith-comparison call depth)
  (reconstruct-arith-binop call depth)
  type/boolean)

(define (reconstruct-arith-monop call depth)
  (let ((id (primop-id (call-exp-proc call)))
	(args (call-exp-args call)))
    (if (not (= 0 (call-exp-exits call)))
	(bug "who's making calls with real exits? ~S" call))
    (let ((type (type-check (car args) depth #f)))
      (make-constrained-arith-monop-type id type args depth call))))

(define (reconstruct-int-binop call depth)
  (let ((args (call-exp-args call)))
    (check-arg-type args 0 type/int32 depth call)
    (check-arg-type args 1 type/int32 depth call)
    type/int32))

(define (reconstruct-int-monop call depth)
  (let ((args (call-exp-args call)))
    (check-arg-type args 0 type/int32 depth call)
    type/int32))

(define-local-syntax (define-c-arith-binop-generator id c-op)
  `(define-c-generator ,id #t
     (lambda (args)
       (,id (car args) (cadr args)))
     (lambda (call depth)
       (reconstruct-arith-binop call depth))
     (lambda (call port indent)
       (simple-c-primop ,c-op call port))))

(define-c-arith-binop-generator + "+")
(define-c-arith-binop-generator - "-")
(define-c-arith-binop-generator * "*")
(define-c-arith-binop-generator quotient  "/")

(define-local-syntax (define-c-int-binop-generator id c-op)
  `(define-c-generator ,id #t
     (lambda (args)
       (,id (car args) (cadr args)))
     (lambda (call depth)
       (reconstruct-int-binop call depth))
     (lambda (call port indent)
       (simple-c-primop ,c-op call port))))

(define-c-int-binop-generator remainder "%")
(define-c-int-binop-generator bitwise-and "&")
(define-c-int-binop-generator bitwise-ior "|")
(define-c-int-binop-generator bitwise-xor "^")

(define-c-generator ashl #t
  (lambda (args)
    (arithmetic-shift (car args) (cadr args)))
  (lambda (call depth)
    (reconstruct-int-binop call depth))
  (lambda (call port indent)
    (generate-shift call port indent "LEFT" "<<")))

(define-c-generator ashr #t
  (lambda (args)
    (arithmetic-shift (car args) (- (cadr args))))
  (lambda (call depth)
    (reconstruct-int-binop call depth))
  (lambda (call port indent)
    (generate-shift call port indent "RIGHT" ">>")))

(define (generate-shift call port indent macro c-op)
  (cond ((= 1 (call-exits call))
	 ; PS_SHIFT_??? is a C macro that handles overshifting even if C doesn't
	 (indent-to port indent)
	 (format port "PS_SHIFT_~A(" macro)
	 (c-value (call-arg call 1) port)
	 (format port ", ")
	 (c-value (call-arg call 2) port)
	 (format port ", ")
	 (c-variable (car (lambda-variables (call-arg call 0))) port)
	 (format port ")"))
	((>= (literal-value (call-arg call 1)) 32)
	 (format port "0L"))
	(else
	 (simple-c-primop c-op call port))))
    
(define-c-generator bitwise-not #t
  (lambda (args)
    (not (car args)))
  (lambda (call depth)
    (reconstruct-int-monop call depth))
  (lambda (call port indent)
    (simple-c-primop "~" call port)))

(define-local-syntax (define-c-comp-binop-generator id c-op arg-type)
  `(define-c-generator ,id #t
     (lambda (args)
       (,id (car args) (cadr args)))
     (lambda (call depth)
       (reconstruct-arith-comparison call depth))
     (lambda (call port indent)
       (simple-c-primop ,c-op call port))))

(define-c-comp-binop-generator =      "==" type/int32)
(define-c-comp-binop-generator <      "<"  type/int32)
(define-c-comp-binop-generator char=? "==" type/int8u)
(define-c-comp-binop-generator char<? "<"  type/int8u)

(define (check-binop-arg-types call type depth)
  (check-call-arg-type call 0 type depth)
  (check-call-arg-type call 1 type depth))

(define-c-generator ascii->char #t
  (lambda (args)
    (ascii->char (car args)))
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/int32 depth call) ;hmm
    type/int8u)
  (lambda (call port indent)
    (c-value (call-arg call 0) port)))
    
(define-c-generator char->ascii #t
  (lambda (args)
    (char->ascii (car args)))
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/int8u depth call)
    type/int8u)
  (lambda (call port indent)
    (c-value (call-arg call 0) port)))
    

