
(define (write-c-io-call call port name . args)
  (format port name)
  (writec port #\()
  (for-each (lambda (arg)
	      (if (string? arg)
		  (format port arg)
		  (c-value (call-arg call arg) port)))
	    args)
  (writec port #\)))

;(define-c-generator cast-to-long #t
;  (lambda (args)
;    (bug "cast-to-long found in source"))
;  (lambda (call depth)
;    (check-call-arg-type call 0 type/int32 depth)
;    type/int32)
;  (lambda (call port indent)
;    (writec port #\()
;    (format port '"(long)")
;    (c-value (call-arg call 0) port)
;    (writec port #\))))

; stdin and stdout cannot be variables because they cannot be declared as
; they may be macros in C.

(define-c-generator stdin #t
  (lambda (args)
    (current-input-port))
  (lambda (call depth)
    type/port)
  (lambda (call port indent)
    (format port "stdin")))

(define-c-generator stdout #t
  (lambda (args)
    (current-output-port))
  (lambda (call depth)
    type/port)
  (lambda (call port indent)
    (format port "stdout")))

;{ int TTchar;
;  PS_GETC(port, TTchar);
;  if (EOF == TTchar) {
;   eof cont }
;  else {
;   cont_var = TTchar;
;   char cont}}

(define-c-generator read-char #f
  (lambda (exits args rep env)
    (eval-call exits args env
	       (lambda (args)
		 (let* ((vals (evaluate-list args env))
			(ch (read-char (caddr vals))))
		   (let ((ch (read-char (caddr args))))
		     (if (eof-object? ch)
			 (apply-closure (cadr args) '())
			 (apply-closure (car args) (list ch))))))))
  (lambda (call depth)
    (reconstruct-read-char call depth))
  (lambda (call port indent)
    (destructure ((#(char-cont eof-cont port-arg) (call-args call)))
      ; PS_GETC is a C macro that may do clearerr after getting an EOF
      (indent-to port indent)
      (format port "{ int TTchar;~%")
      (indent-to port (+ indent 2))
      (format port "PS_GETC(")
      (c-value port-arg port)
      (format port ", TTchar);~%")
      (indent-to port (+ indent 2))
      (format port "if (EOF == TTchar) {~%")
      (write-c-block (lambda-body eof-cont) port (+ indent 4))
      (newline port)
      (indent-to port (+ indent 2))
      (format port "else {~%")
      (indent-to port (+ indent 4))
      (format port "unsigned char ")
      (c-variable (car (lambda-variables char-cont)) port)
      (format port " = TTchar;")
      (write-c-block (lambda-body char-cont) port (+ indent 4))
      (format port "}"))))

(define (reconstruct-read-char call depth)
  (let ((conts+test (call-exp-args call)))
    (if (not (= 2 (call-exp-exits call)))
	(bug "wrong number of exits in a call to READ-CHAR"))
    (let ((port-type (type-check (caddr conts+test) depth (cddr conts+test)))
	  (char-cont-type (type-check-cont (car conts+test) depth))
	  (eof-cont-type  (type-check-cont (cadr conts+test) depth)))
      (unify! port-type type/port call)
      (unify! char-cont-type eof-cont-type call)
      (maybe-follow-uvar char-cont-type))))

(define-c-generator char-ready? #t
  (lambda (args)
    (char-ready? (car args)))
  (lambda (call depth)
    (check-call-arg-type call 0 type/port depth)
    type/boolean)
  (lambda (call port indent)
    (write-c-io-call call port "char_ready_p" 0)))

(define-c-generator write-char #t
  (lambda (args)
    (write-char (car args) (cadr args))
    the-undefined-value)
  (lambda (call depth)
    (check-call-arg-type call 0 type/int8u depth)
    (check-call-arg-type call 1 type/port depth)
    type/unit)
  (lambda (call port indent)
    (write-c-io-call call port "putc" 0 ", " 1)))

(define-c-generator write-string #t
  (lambda (args)
    (display (car args) (cadr args))
    the-undefined-value)
  (lambda (call depth)
    (check-call-arg-type call 0 type/char-pointer depth)
    (check-call-arg-type call 1 type/port depth)
    type/unit)
  (lambda (call port indent)
    (write-c-io-call call port "fputs" 0 ", " 1)))

(define-c-generator force-output #t
  (lambda (args)
    ((structure-ref primitives force-output) (car args))
    the-undefined-value)
  (lambda (call depth)
    (check-call-arg-type call 0 type/port depth)
    type/unit)
  (lambda (call port indent)
    (write-c-io-call call port "fflush" 0)))

; (read-block port buffer count)

(define-c-generator read-block #t
  (lambda (args)
    (bug "no eval method for READ-BLOCK"))
  (lambda (call depth)
    (check-call-arg-type call 0 type/port depth)
    (check-call-arg-type call 1 type/int32 depth)
    (check-call-arg-type call 2 type/int32 depth)
    type/int32)
  (lambda (call port indent)
    (write-c-io-call call port
		     "fread"
		     "(void *)" 1
		     ", sizeof(char), " 2
		     ", " 0)))

; (write-block port buffer count)

(define-c-generator write-block #t
  (lambda (args)
    (bug "no eval method for WRITE-BLOCK"))
  (lambda (call depth)
    (check-call-arg-type call 0 type/port depth)
    (check-call-arg-type call 1 type/int32 depth)
    (check-call-arg-type call 2 type/int32 depth)
    type/int32)
  (lambda (call port indent)
    (write-c-io-call call port
		     "fwrite"
		     "(void *)" 1
		     ", sizeof(char), " 2
		     ", " 0)))

; (open-file name mode)

(define-c-generator open-file #t
  (lambda (args)
    (case (cadr args)
      ((read)
       (open-input-file (car args)))
      ((write)
       (open-output-file (car args)))))
  (lambda (call depth)
    (check-call-arg-type call 0 type/char-pointer depth)
    type/port)
  (lambda (call port indent)
    (write-c-io-call call port
		     "ps_open" 0 ", "
		     (case (literal-value (call-arg call 1))
		       ((read) "\"r\"")
		       ((write) "\"w\"")
		       (else
			(error "bad OPEN specification ~S"
			       (literal-value (call-arg call 1))))))))

(define-c-generator close-port #t
  (lambda (args)
    (if (input-port? (car args))
	(close-input-port (car args))
	(close-output-port (car args))))
  (lambda (call depth)
    (check-call-arg-type call 0 type/port depth)
    type/unit)
  (lambda (call port indent)
    (write-c-io-call call port "fclose" 0)))

(define-c-generator runtime #t
  (lambda (args)
    (bug "no eval method for RUNTIME"))
  (lambda (call depth)
    type/int32)
  (lambda (call port indent)
    (format port "ps_runtime()")))

(define-c-generator abort #t
  (lambda (args)
    (bug "no eval method for ABORT"))
  (lambda (call depth)
    type/unit)
  (lambda (call port indent)
    (format port "(exit -1)")))

(define-c-generator error #f
  (lambda (args)
    (bug "no eval method for ERROR"))
  (lambda (call depth)
    (check-call-arg-type call 0 type/char-pointer depth)
    (check-call-arg-type call 1 type/int8u depth)
    (check-call-arg-type call 2 type/int32 depth)
    (check-call-arg-type call 3 type/int32 depth)
    (check-call-arg-type call 4 type/int32 depth)
    type/null)
  (lambda (call port indent)
    (indent-to port indent)
    (write-c-io-call call port "TTerror" 1 ", " 2 ", " 3 ", " 4 ", " 5)
    (write-char #\; port)))

(define-c-generator null-port? #t
  (lambda (args)
    (bug "no interpreted method for NULL-PORT?"))
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/port depth call)
    type/boolean)
  (lambda (call port indent)
    (display "NULL == " port)
    (c-value (call-arg call 0) port)))