
; (make-vector size init)

(define-c-generator make-vector #t
  (lambda (args)
    (make-stob (make-vector (car args))))
  (lambda (call depth)
    (let* ((var (make-uvar 'vector depth))
	   (type (make-pointer-type var))
	   (args (call-exp-args call)))
      (make-nonpolymorphic! var)       ; avoid polymorphic references
      (if (not (null? (cdr args)))
	  (error "type inferred twice for call ~S" call))
      (set-cdr! args (list (make-quote-exp type #f))); save for future reference
      (check-arg-type args 0 type/int32 depth call)
      type))
  (lambda (call port indent)
    (let ((type (node-type call)))
      (write-c-coercion type port)
      (format port "malloc(sizeof(")
      (display-c-type (pointer-type-to type) #f port)
      (format port ") * ")
      (c-value (call-arg call 0) port)
      (format port ")"))))

(define-c-generator vector-length #t
  (lambda (args)
    (vector-length (de-stob (car args))))
  (lambda (call depth)
    (user-error "vector-length is not available in Pre-Scheme"))
  (lambda (call port indent)
    (user-error "vector-length is not available in Pre-Scheme")))

(define (de-stob x)
  (if (stob? x)
      (stob-value x)
      x))

(define-c-generator vector-ref #t
  (lambda (args)
    (vector-ref (de-stob (car args)) (cadr args)))
  (lambda (call depth)
    (let* ((id (or (find-some-id (car (call-exp-args call))) 'vector-ref))
	   (uvar (make-uvar id depth)))
      (check-call-arg-type call 0 (make-pointer-type uvar) depth)
      (check-call-arg-type call 1 type/int32 depth)
      uvar))
  (lambda (call port indent)
    (generate-c-vector-ref (call-arg call 0) (call-arg call 1) port)))

(define (generate-c-vector-ref vector index port)
  (display "*(" port)
  (c-value vector port)
  (display " + " port)
  (c-value index port)
  (writec port #\)))

(define-c-generator vector-set! #t
  (lambda (args)
    (vector-set! (de-stob (car args)) (cadr args) (caddr args)))
  (lambda (call depth)
    (let* ((id (or (find-some-id (car (call-exp-args call))) 'vector-set!))
	   (uvar (make-uvar id depth)))
      (check-call-arg-type call 0 (make-pointer-type uvar) depth)
      (check-call-arg-type call 1 type/int32 depth)
      (check-call-arg-type call 2 uvar depth)
      type/unit))
  (lambda (call port indent)
    (generate-c-vector-set (call-arg call 1)
			   (call-arg call 2)
			   (call-arg call 3)
			   port indent)))

(define (generate-c-vector-set vector index value port indent)
  (indent-to port indent)
  (generate-c-vector-ref vector index port)
  (display " = " port)
  (c-value value port)
  (writec port #\;))

(define-c-generator make-string #t
  (lambda (args)
    (make-stob (make-string (car args))))
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/int32 depth call)
    type/char-pointer)
  (lambda (call port indent)
    ; calloc is used as a hack to get a zero at the end
    (format port "(unsigned char *)calloc( 1, 1 + ")
    (c-value (call-arg call 0) port)
    (format port ")")))

(define-c-generator string-length #t
  (lambda (args)
    (string-length (de-stob (car args))))
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/char-pointer depth call)
    type/int32)
  (lambda (call port indent)
    (format port "strlen(")
    (c-value (call-arg call 0) port)
    (format port ")")))

(define-c-generator string-ref #t
  (lambda (args)
    (string-ref (de-stob (car args)) (cadr args)))
  (lambda (call depth)
    (check-call-arg-type call 0 type/char-pointer depth)
    (check-call-arg-type call 1 type/int32 depth)
    type/int8u)
  (lambda (call port indent)
    (generate-c-vector-ref (call-arg call 0) (call-arg call 1) port)))

(define-c-generator string-set! #f
  (lambda (args)
    (string-set! (de-stob (car args)) (cadr args) (caddr args)))
  (lambda (call depth)
    (check-call-arg-type call 0 type/char-pointer depth)
    (check-call-arg-type call 1 type/int32 depth)
    (check-call-arg-type call 2 type/int8u depth)
    type/unit)
  (lambda (call port indent)
    (generate-c-vector-set (call-arg call 1)
			   (call-arg call 2)
			   (call-arg call 3)
			   port indent)))

(define-c-generator allocate-memory #t
  (lambda (args)
    (error "no eval method for allocate-memory"))
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/int32 depth call)
    type/int32)
  (lambda (call port indent)
    (write-c-coercion type/int32 port)
    (format port "malloc(") 
    (c-value (call-arg call 0) port)
    (format port ")")))

(define-c-generator deallocate #t
  (lambda (args)
    #f)  ; there must be something better
  (lambda (call depth)
    (let ((var (make-uvar 'deallocate depth)))
      (check-arg-type (call-exp-args call) 0 (make-pointer-type var) depth call)
      type/unit))
  (lambda (call port indent)
    (format port "free(") 
    (c-value (call-arg call 0) port)
    (format port ")")))

(define-c-generator deallocate-memory #t
  (lambda (args)
    #f)  ; there must be something better
  (lambda (call depth)
    (check-arg-type (call-exp-args call) 0 type/int32 depth call)
    type/unit)
  (lambda (call port indent)
    (format port "free(") 
    (c-value (call-arg call 0) port)
    (format port ")")))

;(define-c-generator pointer-add #t
;  (lambda (args)
;    (pointer-add (car args) (cadr args)))
;  (lambda (call depth)
;    (let* ((args (call-exp-args call))
;           (type (type-check (car args) depth args)))
;      (unify! type   ; force a pointer type
;              (make-pointer-type (make-uvar 'vector depth))
;              call)
;      (check-arg-type (call-exp-args call) 1 type/int32 depth call)
;      type))  ; pass on the original, to be referenced or set
;  (lambda (call port indent)
;    (simple-c-primop "+" call port)))

;(define-c-generator string-length #t
;  (lambda (args)
;    (string-length (car args)))
;  (lambda (call depth)
;    (check-call-arg-type call 0 type/pointer depth)
;    type/int32)
;  (lambda (call port indent)
;    (display "(long)strlen((char *)" port)
;    (c-value (call-arg call 0) port)
;    (writec port '#\))))

(define-c-generator computed-goto #f
  (lambda (exits args depth exp)
    (bug "primop/computed-goto found in initial expression"))
  (lambda (call depth)
    (bug "primop/computed-goto found in initial expression"))
  (lambda (call port indent)
    (generate-c-switch call port indent)))

(define (generate-c-switch call port indent)
  (let ((size (call-exits call)))
    (indent-to port indent)
    (display "switch (" port)
    (c-value (call-arg call size) port)
    (display ") {" port)
    (let ((indent (+ indent 2)))
      (do ((i 0 (+ i 1)))
	  ((>= i size))
	(indent-to port indent)
	(format port "case ~D : " i)
	(if (or (= i (- size 1))
		(not (node-equal? (call-arg call i) (call-arg call (+ i 1)))))
	    (write-c-switch-case (call-arg call i) port indent))))
    (indent-to port indent)
    (display "}" port)))

(define (write-c-switch-case node port indent)
  (writec port #\{)
  (write-c-block (lambda-body node) port (+ indent 2))
  (indent-to port (+ indent 2))
  (display "break;" port))
