
(define scheme-primop-table (make-symbol-table))

(walk-vector (lambda (primop)
	       (if (primop? primop)
		   (table-set! scheme-primop-table
			       (primop-id primop)
			       primop)))
	     all-primops)

(define (get-scheme-primop id)
  (cond ((table-ref scheme-primop-table id)
	 => identity)
	((name->enumerand id primop)
	 => get-primop)
	(else
	 (bug "Scheme primop ~A not found" id))))

(define (define-primitive-expander name nargs proc)
  (define-pre-scheme! name nargs
    (lambda (exp env)
      (if (= (length (cdr exp)) nargs)
	  (proc exp (cdr exp) env)
	  (error "wrong number of arguments to primitive ~S" exp)))))

(define (make-primop-visible! id nargs . maybe-primop-or-id)
  (let ((primop (cond ((null? maybe-primop-or-id)
		       (get-scheme-primop id))
		      ((symbol? (car maybe-primop-or-id))
		       (get-scheme-primop (car maybe-primop-or-id)))
		      (else
		       (car maybe-primop-or-id)))))
    (define-primitive-expander id nargs
      (lambda (source args env)
	(make-call-exp primop 0 type/unknown
		       (map (lambda (arg)
			      (expand arg env #f))
			    args)
		       source)))))

(define-syntax define-scheme-primop
  (syntax-rules ()
    ((define-scheme-primop id side-effects type simplifier)
     (table-set! scheme-primop-table 'id
		 (make-primop 'id #t side-effects simplifier
			      (lambda (call) 1)
			      type)))))

(define-syntax define-nonsimple-scheme-primop
  (syntax-rules ()
    ((define-nonsimple-scheme-primop id side-effects simplifier)
     (table-set! scheme-primop-table 'id
		 (make-primop 'id #f side-effects simplifier
			      (lambda (call) 1)
			      'nontrivial-primop)))))

(define-syntax define-scheme-cond-primop
  (syntax-rules ()
    ((define-scheme-cond-primop id simplifier expand simplify?)
     (table-set! scheme-primop-table 'id
		 (make-conditional-primop 'id
					  #f
					  simplifier
					  (lambda (call) 1)
					  expand
					  simplify?)))))

(define-pre-scheme! 'error 4  ; all four args must be present if used as value
  (lambda (exp env)
    (let ((string (expand (cadr exp) env #f))
	  (args (map (lambda (arg)
		       (expand arg env #f))
		     (cddr exp))))
      (make-block-exp
       (list
	(make-call-exp (get-scheme-primop 'error)
		       0
		       type/unknown
		       `(,string
			 ,(make-quote-exp (length args) type/int32)
			 . ,(case (length args)
			      ((0)
			       (list (make-quote-exp 0 type/int32)
				     (make-quote-exp 0 type/int32)
				     (make-quote-exp 0 type/int32)))
			      ((1)
			       (list (car args)
				     (make-quote-exp 0 type/int32)
				     (make-quote-exp 0 type/int32)))
			      ((2)
			       (list (car args)
				     (cadr args)
				     (make-quote-exp 0 type/int32)))
			      ((3)
			       args)
			      (else
			       (error "too many arguments to ERROR in ~S" exp))))
		       exp)
	(make-quote-exp the-undefined-value type/unknown))))))
