

(define-scheme-primop make-vector 'allocate
  (lambda (call)
    (literal-value (call-arg call 1)))  ; stashed here during inference
  default-simplifier)

(define-scheme-primop vector-length #f  ; only for initial evaluation
  (lambda (call) type/int32)
  default-simplifier)

(define-scheme-primop vector-ref 'read
  (lambda (call)
    (pointer-type-to (node-type (call-arg call 0))))
  default-simplifier)

(define-nonsimple-scheme-primop vector-set! 'write default-simplifier)

(define-scheme-primop make-string 'allocate
  (lambda (call) type/char-pointer)
  default-simplifier)

(define-scheme-primop string-length #f
  (lambda (call) type/int32)
  default-simplifier)

(define-scheme-primop string-ref 'read
  (lambda (call) type/int8u)
  default-simplifier)

(define-nonsimple-scheme-primop string-set! 'write default-simplifier)

(define-scheme-primop deallocate 'deallocate
  (lambda (call) type/unit)
  default-simplifier)

(define-scheme-primop allocate-memory 'allocate
  (lambda (call) type/int32)
  default-simplifier)

(define-scheme-primop deallocate-memory 'deallocate
  (lambda (call) type/unit)
  default-simplifier)

(define-primitive-expander 'contents-int8u 1
  (lambda (source args cenv)
    (dereference-cell source (expand (car args) cenv #f) type/int8u 'memory)))

(define-primitive-expander 'set-contents-int8u! 2
  (lambda (source args cenv)
    (destructure (((pointer value) args))
      (set-cell source
		(expand pointer cenv #f)
		(expand value cenv #f)
		type/int8u
		'memory))))

(define-primitive-expander 'contents-int32 1
  (lambda (source args cenv)
    (dereference-cell source (expand (car args) cenv #f) type/int32 'memory)))

(define-primitive-expander 'set-contents-int32! 2
  (lambda (source args cenv)
    (destructure (((pointer value) args))
      (set-cell source
		(expand pointer cenv #f)
		(expand value cenv #f)
		type/int32
		'memory))))

(for-each (lambda (p)
	    (make-primop-visible! (car p) (cadr p)))
	  '((make-vector       1)
	    (vector-length     1)
	    (vector-ref        2)
	    (vector-set!       3)
	    (make-string       1)
	    (string-length     1)
	    (string-ref        2)
	    (string-set!       3)
	    (deallocate        1)
	    (allocate-memory   1)
	    (deallocate-memory 1)
	    ))

; (COMPUTED-GOTO <exit0> <exit1> ... <exitN> <dispatch-value>)

; Remove an unecessary coercion on the disptach-value, if possible.

(define (simplify-computed-goto call)
  (simplify-args call 0)
  (let ((value (call-arg call (call-exits call))))
    (cond ((and (call-node? value)
		(eq? 'coerce (primop-id (call-primop value)))
		(< (call-exits call) 256)
		(eq? type/int8u (literal-value (call-arg value 1))))
	   (replace value (detach (call-arg value 0)))))))

(define-nonsimple-scheme-primop computed-goto #f simplify-computed-goto)

