
; Entry point

(define (pre-scheme package-id spec-files init-name c-file . commands)
  (reset-node-id)
  (receive (no-copy shadow integrate)
      (parse-prescheme-commands commands)
    (receive (scanner pre-scheme exports lookup)
	(package-specs->scanner+exports package-id spec-files)
      (let ((forms (flatten-forms (scan-making-forms scanner pre-scheme))))
	(mark-forms! (list (cons package-id exports))
		     lookup
		     (lambda (f) (set-form-exported?! f #t))
		     "exported")
	(mark-forms! no-copy
		     lookup
		     (lambda (f) (set-form-integrate! f 'no))
		     "not to be copied")
	(reset-type-vars!)
	(receive (forms producer)
	    (sort-forms forms)
	  (let ((sorted (let loop ((forms '()))
			     (cond ((producer)
				    => (lambda (f)
					 (type-check-form f)
					 (loop (cons f forms))))
				   (else
				    (reverse forms))))))
	    (add-type-coercions)
	    (format #t "Simplifying~%")
	    (for-each (lambda (form)
			(format #t " ~A " (form-name form))
			(let ((status (expand-and-simplify-form form)))
			  (if status
			      (format #t "(~A): " status)
			      (format #t ": "))
			  (display-type (variable-type (form-var form))
					(current-output-port))
			  (newline)))
		      sorted)
	    (format #t "Procedures simplified~%")
	    (let ((forms (remove-unreferenced-forms forms)))
	      (for-each integrate-stob-form forms) 
	      ; prevent further automatic integration
	      (for-each (lambda (form)
			  (remove-known-value (form-var form)))
			forms)
	      (integrate-by-command integrate forms)
	      (for-each resimplify-form forms)
	      (let* ((forms (remove-unreferenced-forms forms))
		     (forms (integrate-single-uses forms))
		     (forms (remove-unreferenced-forms forms)))
		(for-each resimplify-form forms)
		(for-each remove-polymorphism forms)
					; (if cps-file (write-cps-file cps-file forms))
		(if c-file (write-c-file init-name c-file forms shadow))))))))))
  
(define (mark-forms! specs lookup marker mark)
  (let ((lose (lambda (p n)
		(format #t "Warning: no definition for ~S, cannot mark as ~A~%"
			(list p n) mark))))
    (for-each (lambda (spec)
		(let ((package-id (car spec))
		      (ids (cdr spec)))
		  (for-each (lambda (id)
			      (let ((var (lookup package-id id)))
				(cond ((and (variable? var)
					    (variable-form var))
				       => marker)
				      (else
				       (lose package-id id)))))
			    ids)))
	      specs)))

(define (parse-prescheme-commands commands)
  (let loop ((commands commands) (no-copy '()) (shadow '()) (integrate '()))
    (if (null? commands)
	(values no-copy shadow (reverse integrate))
	(let ((key (caar commands))
	      (more (cdar commands))
	      (rest (cdr commands)))
	  (case key
	    ((no-copy)
	     (loop rest (append more no-copy) shadow integrate))
	    ((shadow)
	     (loop rest no-copy (append more shadow) integrate))
	    ((integrate)
	     (loop rest no-copy shadow (append (reverse more) integrate)))
	    (else
	     (error "unknown directive ~S" (cons key more))))))))

(define (scan-making-forms scanner pre-scheme)
  (let* ((index 0)
	 (next-index (lambda ()
		       (let ((i index))
			 (set! index (+ index 1))
			 i)))
	 (forms '()))
    (allow-references-to-primitives
     (lambda (id nargs)
       (receive (ignore value free-vars)
	   (expand-node (make-primitive-proc id nargs pre-scheme) pre-scheme)
	 (let* ((var (make-global-variable id type/unknown))
		(form (make-form var (evaluate value) #f (next-index) free-vars)))
	   (set! forms (cons form forms))
	   var)))
     (lambda ()
       (scanner
	(lambda (package nodes)
	  (for-each
	   (lambda (node)
	     (receive (var value free-vars)
		 (expand-node node package)
	       (let ((value (evaluate value)))
		 (if var
		     (let* ((source (cadr (schemify node)))
			    (form (make-form var value source (next-index) free-vars)))
		       (set! forms (cons form forms)))))))
	   nodes)))))
    forms))

(define (make-primitive-proc id nargs package)
  (let ((args (do ((nargs nargs (- nargs 1))
		   (args '() (cons (string->symbol (format #f "x~D" nargs))
				   args)))
		  ((<= nargs 0) args))))
    ((structure-ref syntactic classify)
     `(lambda ,args
	(,id . ,args))
     ((structure-ref packages package->environment) package))))

(define (flatten-forms forms)
  (let loop ((forms forms) (all-forms '()))
    (if (null? forms)
	(reverse all-forms)
	(let ((form (car forms)))
	  (receive (value free more-forms)
	      (flatten (form-var form) (form-value form) (form-free form))
	    (set-form-value! form value)
	    (set-form-free! form free)
	    (loop (cdr forms) (append more-forms (list form) all-forms)))))))

; Three possibilities:
; 2. The variable is settable but the thunk gives it no particular value.
; 3. A real value is or needs to be present, so we relate the type of
; the variable with the type of the value.

; thunk's value may be a STOB and not a lambda.

(define (type-check-form form)
  ;; (format #t "  ~S: " (variable-name (form-var form)))
  (let* ((value (form-value form))
	 (var (form-var form))
	 (name (form-name form))
	 (value-type (cond ((closure? value)
			    (type-check-exp (closure-exp value) value name))
			   ((stob? value)
			    (type-check-stob value name))
			   ((variable? value)
			    (type-check-exp value form name))
			   (else
			    (let ((value (make-quote-exp value type/unknown)))
			      (set-form-value! form value)
			      (type-check-exp value form name))))))
    (cond ((not (variable-settable? var))
	   (let ((type (cond ((eq? type/unknown (variable-type var))
			      (let ((type (schemify-type value-type 0)))
				(set-variable-type! var type)
				type))
			     (else
			      (unify! value-type (get-variable-type var) form)
			      value-type))))
	     (if (not (type-scheme? type))
		 (make-nonpolymorphic! type)) ; lock down any related uvars
	     ;;(format #t "~S~%" (instantiate type))
	     ))
	  ((not (or (eq? type/unit value-type)
		    (eq? type/null value-type)))
	   (make-nonpolymorphic! value-type) ; no polymorphism allowed (so it
	   ;; is not checked for, so there may be depth 0 uvars in the type)
	   ;; (format #t " ~S~%" (instantiate value-type))
	   (unify! value-type (get-variable-type var) form))
	  (else
	   0 ; prevent empty block warning
	   ;;(format #t " ~S~%" (instantiate value-type))
	   ))))

;--------------------------------------------------

(define (integrate-single-uses forms)
  (format #t "In-lining single-use procedures~%")
  (filter (lambda (f)
	    (let ((var (form-var f)))
	      (cond ((and (not (form-exported? f))
			  (eq? (form-type f) 'lambda)
			  (not (null? (variable-refs var)))
			  (null? (cdr (variable-refs var)))
			  (procedure-node? (car (variable-refs var))))
		     (format #t " ~S~%" (variable-name (form-var f)))
		     (integrate-single-use
		      f
		      (car (variable-refs (form-var f)))
		      #f)
		     (set-form-type! f 'unused)
		     (set-form-node! f #f '())
		     #f)
;		    ((eq? (form-name f) 'setter)
;		     (breakpoint "integrate-single-use")
;		     #t)
		    (else
		     #t))))
	  forms))

(define (integrate-single-use form ref copy?)
  (let* ((in-form (node-flag (node-base ref)))
	 (in-node (form-value in-form))
	 (type (variable-type (form-var form))))
    (make-form-current in-form)
    (let ((node (cond (copy?
		       (copy-node-tree (form-value form)))
		      (else
		       (add-lambdas (form-lambdas form))
		       (form-value form)))))
      (if (type-scheme? type)
	  (if (not (called-node? ref))
	      (error "integrating polymorphic value into non-call position")
	      (instantiate-type&value type node ref)))
      (determine-lambda-protocol node (list ref))
      (replace ref node)
      (simplify-node in-node)
      (set-node-flag! in-node #f)
      (set-form-node! in-form in-node (make-lambda-list)))))

; Commands are (<proc> <caller>)

(define (integrate-by-command commands forms)
  (for-each (lambda (command)
	      (receive (proc refs)
		  (process-integrate-command command forms)
		(if proc
		    (for-each (lambda (r)
				(integrate-single-use proc r #t))
			      refs))))
	    commands))

; Horrendous error checking and notification.

(define (process-integrate-command command forms)
  (let* ((proc (any (lambda (f)
		      (eq? (form-name f) (car command)))
		    forms))
	 (var (if proc (form-var proc) #f))
	 (node (if proc (form-value proc) #f))
	 (caller (any (lambda (f)
			(eq? (form-name f) (cadr command)))
		      forms))
	 (refs (if (and var caller)
		   (filter (lambda (ref)
			     (eq? caller (node-flag (node-base ref))))
			   (variable-refs var))
		   #f)))
    (cond ((or (not proc) (not var) (not caller))
	   (cond ((or (not proc) (not var))
		  (format #t "Bad command: no value for ~S~%"
			  (car command)))
		 ((or (not node)
		      (not (lambda-node? node)))
		  (format #t "Bad command: ~S is not a procedure~%"
			  (car command))))
	   (if (not caller)
	       (format #t "Bad command: no definition for ~S~%"
		       (car command)))
	   (values #f #f))
	  ((or (null? refs) (not node) (not (lambda-node? node)))
	   (if (null? refs)
	       (format #t "Bad command: ~S is not referenced by ~S~%"
		       (car command) (cadr command)))
	   (if (or (not node)
		   (not (lambda-node? node)))
	       (format #t "Bad command: ~S is not a procedure~%"
		       (car command)))
	   (values #f #f))
	  (else
	   (values proc refs)))))

;--------------------------------------------------
; Determine an actual type for a polymorphic procedure.

(define (remove-polymorphism form)
  (if (type-scheme? (variable-type (form-var form)))
      (make-monomorphic! (form-var form))))

;--------------------------------------------------

; Various methods for getting values from thunks.  These are no longer used
; here.

(define (thunk-value thunk)
  (let ((refs (variable-refs (car (lambda-variables thunk)))))
    (if (= 1 (length refs))
        (call-arg (node-parent (car refs)) 2)
        #f)))

(define (simple-thunk? thunk value)
  (eq? (node-parent (node-parent value)) thunk))

