
; Type schemes

(define-record-type type-scheme
  (
   type           ; a type
   free-uvars     ; uvars that are free and unconstrained
   lattice-uvars  ; uvars that are free but constrained
   )
  ())

(define make-type-scheme type-scheme-maker)

(define-record-discloser type/type-scheme
  (lambda (type-scheme)
    (list 'type-scheme
	  (map uvar-id (type-scheme-free-uvars type-scheme))
	  (type-scheme-type type-scheme))))

(define (simple-polymorphic-scheme? type-scheme)
  (null? (type-scheme-lattice-uvars type-scheme)))

; If TYPE has any variables bound at DEPTH this returns a type scheme making
; those variables polymorphic; otherwise TYPE is returned.

; Would like to do limited finalizing of uvars, but can't.
; Consider (lambda (g x) (tuple (g 3) (g x) x))
; (a -> b) -> c -> [d, e, f] with
; a > int8, d > b, a > c, e > b, f > c
; No polymorphism, and no simplification without restricting someone
; But consider NOT  a ->b, bool > a, b > bool
; It could just as well be bool -> bool.
; Simplification okay on variables that are not used inside other types?

(define *free-uvars* '())
(define *lattice-uvars* '())

(define (schemify-type type depth)
  (let ((overload? #f))
  (set! *free-uvars*    '())
  (set! *lattice-uvars* '())
  (find-free-uvars type depth overload?)
  (let ((lattice-uvars (map (lambda (uvar)
			      (let ((new (uvar-place uvar)))
				(set-uvar-place! uvar #f)
				new))
			    *lattice-uvars*))
	(free-uvars *free-uvars*))
    (set! *free-uvars*    '()) ; drop pointers
    (set! *lattice-uvars* '())
    (for-each (lambda (uvar)
		(set-uvar-place! uvar #f))
	      free-uvars)
    (if (or (not (null? lattice-uvars))
	    (not (null? free-uvars)))
	(make-type-scheme type free-uvars lattice-uvars)
	type))))
  
(define (find-free-uvars type depth overload?)
  (let label ((type type))
    (cond ((other-type? type)
	   (for-each label (other-type-subtypes type)))
	  ((not (uvar? type)))
	  ((uvar-binding type)
	   => label)
	  ((and (not (uvar-place type))
		(<= depth (uvar-depth type)))
	   (find-free-uvar type depth overload?)))))

(define (find-free-uvar uvar depth overload?)
  (cond ((or (not (uvar-constrained? uvar))
	     (not (uvar-constrained-at-depth? uvar depth)))
	 (set-uvar-place! uvar uvar)
	 (set! *free-uvars* (cons uvar *free-uvars*)))))

;	(overload?
;	 (copy-lattice-uvar uvar depth))

(define (uvar-constrained-at-depth? uvar depth)
  (let ((flag (make-vector 1)))
    (define (do-sub relation)
      (do-uvar (relation-source relation)))
    (define (do-super relation)
      (do-uvar (relation-dest relation)))
    (define (do-uvar uvar)
      (cond ((or (not (uvar? uvar))
		 (uvar-arith-loc uvar)
		 (> depth (uvar-depth uvar)))
	     #t)
	    ((eq? flag (uvar-temp uvar))
	     #f)
	    (else
	     (set-uvar-temp! uvar flag)
	     (or (any? do-super (uvar-supers uvar))
		 (any? do-sub   (uvar-subs   uvar))))))
    (cond ((do-uvar uvar) #t)
	  (else
	   (bind-related-uvars! uvar)
	   (set-uvar-constrained?! uvar #f)
	   (set-uvar-subs! uvar '())
	   (set-uvar-supers! uvar '())
	   #f))))

(define (bind-related-uvars! initial-uvar)
  (define (do-uvar uvar)
    (cond ((not (or (uvar-binding uvar)
		    (eq? uvar initial-uvar)))
	   (set-uvar-binding! uvar initial-uvar)
	   (for-each do-relation-source (uvar-subs uvar))
	   (for-each do-relation-dest (uvar-supers uvar))
	   (set-uvar-subs! uvar '())
	   (set-uvar-supers! uvar '()))))

  (define (do-relation-source relation)
    (do-uvar (relation-source relation)))
  (define (do-relation-dest relation)
    (do-uvar (relation-dest relation)))

  (for-each do-relation-source (uvar-subs initial-uvar))
  (for-each do-relation-dest (uvar-supers initial-uvar)))

; This code is completely bogus.

;(define (copy-lattice-uvar uvar depth)
;  (define (copy-relation relation)
;    (make-relation (copy (relation-dest    relation))
;                   (copy (relation-source1 relation))
;                   (copy (relation-source2 relation))
;                   (relation-site          relation)
;                   (relation-transmitter   relation)))
;  (define (copy type)
;    (cond ((not (uvar? type))  ; lattice type
;           type)
;          ((uvar-place uvar)
;           => identity)
;          ((> depth (uvar-depth uvar))
;           uvar)
;          (else
;           (let ((new (make-uvar (uvar-prefix uvar) (uvar-depth uvar))))
;             (set! *lattice-uvars* (cons uvar *lattice-uvars*))
;             (set-uvar-place! uvar new)
;             (make-uvar-constrained! new)
;             (set-uvar-supers! new (map copy-relation (uvar-supers uvar)))
;             (set-uvar-subs!   new (map copy-relation (uvar-subs   uvar)))
;             new))))
;  (copy uvar))
    
; Instantiate SCHEME at DEPTH.
;
; New sequence:
;   (instantiate-type-scheme scheme depth)
;   ... elide bindings in new copy ...
;   (clean-type-scheme scheme)

(define (instantiate-type-scheme scheme depth . maybe-thunk)
  (instantiate-type-scheme! scheme depth)
  (let ((type (copy-type (type-scheme-type scheme))))
    (if (not (null? maybe-thunk))
	((car maybe-thunk)))
    (clean-type-scheme! scheme)
    type))

(define (instantiate-type-scheme! scheme depth)
  (let ((constrained (type-scheme-lattice-uvars scheme))
	(uid (unique-id)))
    (for-each (lambda (uvar)
		(set-uvar-place!
		 uvar
		 (make-uvar (uvar-prefix uvar) depth uid)))
	      (type-scheme-free-uvars scheme))
    (for-each (lambda (uvar)
		(set-uvar-place!
		 uvar
		 (make-uvar (uvar-prefix uvar) depth uid)))
	      constrained)))
;    (for-each instantiate-uvar constrained)

(define (clean-type-scheme! scheme)
  (for-each (lambda (uvar)
	      (set-uvar-place! uvar #f))
	    (type-scheme-free-uvars scheme))
  (for-each (lambda (uvar)
	      (set-uvar-place! uvar #f))
	    (type-scheme-lattice-uvars scheme)))

;(define (instantiate-uvar uvar)
;  (let ((new (uvar-place uvar)))
;    (mark-uvar-constrained! new)
;    (set-uvar-supers! new (map instantiate-relation (uvar-supers uvar)))
;    (set-uvar-subs!   new (map instantiate-relation (uvar-subs   uvar)))))
;
;(define (instantiate-relation relation)
;  (make-relation (instantiate-x (relation-dest    relation))
;                 (instantiate-x (relation-source1 relation))
;                 (instantiate-x (relation-source2 relation))
;                 (relation-site  relation)   ; WRONG - need to make a new copy
;                 (relation-transmitter? relation))) ; of the source as well
;
;(define (instantiate-x x)
;  (cond ((and (uvar? x)
;              (uvar-place x))
;         => identity)
;        (else x)))

(define (copy-type type)
  (cond ((other-type? type)
	 (make-other-type (other-type-kind type)
			  (map copy-type
			       (other-type-subtypes type))))
	((not (uvar? type))
	 type)
	((uvar-binding type)
	 => copy-type)
	((uvar-place type)
	 => identity)
	(else
	 type)))


