
; Type variables - what a mess

(define-record-type uvar
  (prefix        ; a name for debugging
   (depth)       ; lexical depth of the uvar
   id            ; a number
   arith-loc     ; program location if this uvar corresponds to an arithmetic
                 ; operation
   )
  ((place        #f)  ; used in producing type schemes
   (source       #f)  ; to let the user know where this came from

   (binding      #f)  ; known value of this uvar

   (constrained? #f)  ; is this uvar constrained?
   (supers       '()) ; superior relations
   (subs         '()) ; subordinate relations

   (temp         #f)  ; useful field
   ))

(define-record-discloser type/uvar
  (lambda (uvar)
    (list 'uvar
	  (uvar-prefix uvar)
	  (uvar-depth uvar)
	  (uvar-id uvar)
	  (uvar-arith-loc uvar)
	  (uvar-constrained? uvar)
	  (uvar-binding uvar))))

(define (make-uvar prefix depth . maybe-id)
  (uvar-maker prefix
	      depth
	      (if (null? maybe-id)
		  (unique-id)
		  (car maybe-id))
	      #f))

(define (make-arith-uvar prefix arith-loc depth . maybe-id)
  (uvar-maker prefix
	      depth
	      (if (null? maybe-id)
		  (unique-id)
		  (car maybe-id))
	      arith-loc))

(define *constrained-uvars* '())  ; list of all constrained uvars

(define (constrained-uvars)
  (let ((cus (filter uvar-constrained? *constrained-uvars*)))
    (set! *constrained-uvars* cus)
    cus))

(define (mark-uvar-constrained! uvar)
  (cond ((not (uvar-constrained? uvar))
	 (set! *constrained-uvars* (cons uvar *constrained-uvars*))
	 (set-uvar-constrained?! uvar #t))))

(define-record-type relation
  ((source)    ; sources     (uvar or type)
   (dest)      ; destination (uvar or type)
   site        ; where to put the coercion
   )
  (temp))      ; useful slot

;(define make-relation relation-maker)

(define (make-relation source dest site)
  (if (not site)
      (breakpoint "no site for relation ~S => ~S" source dest))
;  (if (call-exp? site)
;      (breakpoint "site is an expression ~S" site))
  (relation-maker source dest site))

(define-record-discloser type/relation
  (lambda (relation)
    (list (relation-source relation)
	  '=>
	  (relation-dest relation))))

(define (add-relation source dest site)
  (let ((new (make-relation source dest site)))
    (cond ((uvar? source)
	   (set-uvar-supers! source (cons new (uvar-supers source)))
	   (mark-uvar-constrained! source)))
    (cond ((uvar? dest)
	   (set-uvar-subs! dest (cons new (uvar-subs dest)))
	   (mark-uvar-constrained! dest)))
    new))
    
; Return a type greater or equal to TYPE, which is the type of SITE.
; Arithmetic operators do not pass in a site as they will be creating
; their own relations.

(define (make-constrained-type prefix type depth site)
  (let ((type (maybe-follow-uvar type)))
    (if (and site
	     (or (uvar? type)
		 (lattice-type? type)))
	(let ((new (make-uvar prefix depth)))
	  (add-relation type new site)
	  new)
	type)))

(define (make-constrained-arith-monop-uvar prefix type loc depth site)
  (let ((type (maybe-follow-uvar type)))
    (if (and (uvar? type)
	     (lattice-type? type))
	(let ((new (make-arith-uvar prefix site depth)))
	  (add-relation type new loc)
	  new)
	#f)))

(define (make-constrained-arith-binop-uvar prefix type1 loc1 type2 loc2 depth site)
  (let ((type1 (maybe-follow-uvar type1))
	(type2 (maybe-follow-uvar type2)))
    (if (and (or (uvar? type1)
		 (lattice-type? type1))
	     (or (uvar? type2)
		 (lattice-type? type2)))
	(let ((new (make-arith-uvar prefix site depth)))
	  (add-relation type2 new loc2) ; to get the source list in the same
	  (add-relation type1 new loc1) ; order as the args
	  new)
	#f)))

; Could this safely short-circuit the chains?

(define (maybe-follow-uvar type)
  (cond ((and (uvar? type)
	      (uvar-binding type))
	 => maybe-follow-uvar)
	(else type)))

; Set the binding of UVAR to be VALUE, merging constraints or propogating
; the binding if necessary.  Returns a lattice type or an arithmetic-op
; uvar (courtesy of PROPOGATE-BINDING) in the event of a type error.

(define (bind-uvar! uvar value)
  (cond ((not (uvar-constrained? uvar))
	 (minimize-type-depth! value (uvar-depth uvar))
	 (set-uvar-binding! uvar value)
	 #f)
	((lattice-type? value)
	 (set-uvar-binding! uvar value)
	 (for-each (lambda (relation)
		     (set-relation-source! relation value))
		   (uvar-supers uvar))
	 (for-each (lambda (relation)
		     (set-relation-dest! relation value))
		   (uvar-subs uvar))
	 #f)
	((uvar? value)       ; move the constraints to VALUE
	 (minimize-type-depth! value (uvar-depth uvar))
	 (set-uvar-binding! uvar value)
	 (merge-constraints! uvar value)
	 #f)
	(else
	 (propogate-binding! uvar value))))

; Make the depths of any uvars in TYPE be no greater than DEPTH.

(define (minimize-type-depth! type depth)
  (let label ((type type))
    (cond ((other-type? type)
	   (for-each label (other-type-subtypes type)))
	  ((uvar? type)
	   (cond ((uvar-binding type)
		  => label)
		 ((< depth (uvar-depth type))
		  (set-uvar-depth! type depth)))))))

; Merge the constraints of UVAR into those of uvar VALUE.

(define (merge-constraints! uvar value)
  (mark-uvar-constrained! value)
  (for-each (lambda (relation)
	      (set-relation-source! relation value))
	    (uvar-supers uvar))
  (set-uvar-supers! value
		    (append (uvar-supers uvar)
			    (uvar-supers value)))
  (for-each (lambda (relation)
	      (set-relation-dest! relation value))
	    (uvar-subs uvar))
  (set-uvar-subs! value
		  (append (uvar-subs uvar)
			  (uvar-subs value))))

; Bind INITIAL-UVAR, which has constraints, to VALUE, which is not a lattice
; type.

(define (propogate-binding! initial-uvar value)
  (call-with-current-continuation
    (lambda (exit)                    ; for returning error values
      (let ((depth (uvar-depth initial-uvar)))

	(define (start)  ; how it all begins
	  (do-uvar initial-uvar)
	  (minimize-type-depth! value depth)
	  #f)

	(define (do-uvar uvar)
	  (cond ((uvar-binding uvar)
		 (do-type (uvar-binding uvar)))
		(else
		 (if (< (uvar-depth uvar) depth)
		     (set! depth (uvar-depth uvar)))
		 (set-uvar-binding! uvar value)
		 (if (not (uvar-arith-loc uvar))
		     (for-each do-relation-source (uvar-subs uvar))
		     (exit uvar))
		 (for-each do-relation-dest (uvar-supers uvar)))))

	(define (do-relation-source relation)
	  (do-type (relation-source relation)))
	(define (do-relation-dest relation)
	  (do-type (relation-dest relation)))

	(define (do-type type)
	  (cond ((type-eq? value type) (values))
		((uvar? type)          (do-uvar type))
		(else                  (exit type))))

	(start)))))

;(define (propogate-binding! initial-uvar value)
;  (let ((depth (uvar-depth initial-uvar)))
;
;    (define (start)  ; how it all begins
;      (do-uvar initial-uvar #f #f)
;      (minimize-type-depth! value depth))
;
;    (define (do-uvar uvar uvar-is-dest? relation)
;      (cond ((uvar-binding uvar)
;             (if (eq? '+ (uvar-arith-op uvar))
;                 (pointer-addition uvar uvar-is-dest? relation))
;             (do-type (uvar-binding uvar) uvar-is-dest? relation))
;            (else
;             (if (< (uvar-depth uvar) depth)
;                 (set! depth (uvar-depth uvar)))        
;             (set-uvar-binding! uvar value)
;             (cond ((not (uvar-arith-loc uvar))
;                    (for-each do-relation-source (uvar-subs uvar)))
;                   ((pointer-type? value)
;                    (case (uvar-arith-op uvar)
;                      ((+) (pointer-addition uvar uvar-is-dest? relation))
;                      ((-) (pointer-subtraction uvar uvar-is-dest? relation))
;                      (else
;                       (type-error "non-arithmetic value passed to arith op"))))
;                   (else
;                    (type-error "non-arithmetic value passed to arith op")))
;             (for-each do-relation-dest (uvar-supers uvar)))))
;
;    ; an error if the pointer type is the second argument, which must be
;    ; an integer
;    (define (pointer-subtraction uvar uvar-is-dest? relation)
;      (let ((rel1 (car (uvar-subs uvar)))
;            (rel2 (cadr (uvar-subs uvar))))
;        (if (and uvar-is-dest?
;                 (eq? relation rel2))
;            (type-error "non-arithmetic value passed to arith op"))
;        (set-relation-dest! rel2 type/int32)
;        ; propogate down if necessary
;        (if (not uvar-is-dest?)
;            (do-relation-source rel1))))
;
;    ; other argument must be an integer
;    (define (pointer-addition uvar uvar-is-dest? relation)
;      (if uvar-is-dest?
;          (let ((rel1 (car (uvar-subs uvar)))
;                (rel2 (cadr (uvar-subs uvar))))
;            (set-relation-dest! (if (eq? relation rel1) rel2 rel1) type/int32))))
;
;    (define (do-relation-source relation)
;      (do-type (relation-source relation) #f relation))
;    (define (do-relation-dest relation)
;      (do-type (relation-dest relation) #t relation))
;
;    (define (do-type type type-is-dest? relation)
;      (cond ((type-eq? value type) (values))
;            ((uvar? type)     (do-uvar type type-is-dest? relation))
;            (else (type-error ...))))
;
;    (start)))

; Set the depth of all uvars in TYPE to be -1 so that it will not be made
; polymorphic at any level.

(define (make-nonpolymorphic! type)
  (cond ((uvar? type)
	 (set-uvar-depth! type -1))
	((other-type? type)
	 (for-each make-nonpolymorphic! (other-type-subtypes type)))
	;((type-scheme? type)
	; (make-nonpolymorphic! (type-scheme-type type)))
	))

;------------------------------------------------------------
; Micro utilities

(define *unique-id-counter* 0)

(define (unique-id)
  (set! *unique-id-counter* (+ *unique-id-counter* 1))
  *unique-id-counter*)

(define (reset-type-vars!)
  (set! *unique-id-counter* 0)
  (set! *constrained-uvars* '()))
