
; Replace a reference to a global cell with the contents of the cell, if the
; contents are known at compile time.  If the contents are not known and the
; cell is not settable, then move the dereferencing call down to the point the
; contents are used.

(define (simplify-global-reference ref)
  (let ((value (variable-known-value (reference-variable ref))))
    (if value
	(let ((node (vector->leaf-node value)))
	  (if node
	      (replace ref node))))))

(define (simplify-unknown-call call)
  (simplify-args call 0)
  (let ((proc (call-arg call 1)))
    (cond ((lambda-node? proc)
	   (determine-lambda-protocol proc (list proc))
	   (mark-changed proc))
	  ((known-value proc)
           => (lambda (value)
		(let ((value (vector->node value)))
		  (if (literal-node? value)
		      (error "calling ~S which has value ~S"
			     (variable-name (call-arg call loc/owner))
			     (literal-value value)))
		  (replace proc value)
		  (mark-changed value)))))))

(define (known-value node)	   
  (if (reference-node? node)
      (variable-known-value (reference-variable node))
      #f))